/*
 * ax25_mac.c
 *
 * MAC layer for AX.25 (AX25_MAC)
 *
 * Parts taken from TFPCX:
 * Copyright (C) 1997  Rene Stange, DG0FT, <stange@berlin.snafu.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*** Include ****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include <string.h>
#include <malloc.h>
#include <io.h>
#include <fcntl.h>
#include "mac_if.h"

/*** Define *****************************************************************/

#define __VER__     "0.05 "     /* version string */
#define VERSION     0x005       /* version number */

#define L1PORTS     8           /* number of L1 ports */
#define MAC_BUFS    100         /* number of buffers */
#define HOSTBUF     512         /* must be a power of 2 */
#define STACK       256         /* words */

/* Ports */
#define PIC0        0x20        /* Master PIC */
#define PIC1        0xA0        /* Slave PIC */
#define PIT         0x40        /* Timer Port 0 */
#define PPI         0x61        /* Parallel Port (Beeper) */

#define EOI         0x20        /* end of interrupt */
#define SETPRIO     0xC0        /* set IRQ priority */

#define IRQ0        0x08        /* int vector of IRQ0 */
#define IRQ8        0x70        /* int vector of IRQ8 */

/* Modem */
#define MODEMS      2

/* PAR */
#define TXD         0x01
#define PTT         0x02
#define BURST       0x04
#define DCDBIT      0x10
#define RXD         0x20

#define IRQEN       0x10

#define DCDCOUNT    5
#define DCDABORT    5
#define DCDLIMIT    15

/* YAM */

/* Offset of UART registers */
#define RXB     0
#define TXB     0
#define DLSB    0
#define DMSB    1
#define IER     1
#define IIR     2
#define LCR     3
#define MCR     4
#define LSR     5
#define MSR     6

/* Interrupt Identification Register Bit Masks */
#define IIR_NOPEND 1
#define IIR_MODEM  0
#define IIR_TX     2
#define IIR_RX     4
#define IIR_LINE   6

/* Interrupt Enable Register Bit Masks */
#define IER_RX  1
#define IER_TX  2
#define IER_LSR 4
#define IER_MSR 8

/* Modem Control Register Bit Masks */
#define MCR_DTR  0x01
#define MCR_RTS  0x02
#define MCR_OUT1 0x04
#define MCR_OUT2 0x08
#define MCR_LOOP 0x10

/* Line Status Register Bit Mask */
#define LSR_RXC    0x01
#define LSR_OE     0x02
#define LSR_PE     0x04
#define LSR_FE     0x08
#define LSR_BREAK  0x10
#define LSR_THRE   0x20
#define LSR_TSRE   0x40

/* Line Control Register Bit Masks */
#define LCR_DLAB    0x80
#define LCR_BREAK   0x40
#define LCR_PZERO   0x28
#define LCR_PEVEN   0x18
#define LCR_PODD    0x08
#define LCR_STOP1   0x00
#define LCR_STOP2   0x04
#define LCR_BIT5    0x00
#define LCR_BIT6    0x02
#define LCR_BIT7    0x01
#define LCR_BIT8    0x03

/* Modem Status Register Bit Masks */
#define MSR_DCTS 0x01
#define MSR_DDSR 0x02
#define MSR_DRIN 0x04
#define MSR_DDCD 0x08
#define MSR_CTS  0x10
#define MSR_DSR  0x20
#define MSR_RING 0x40
#define MSR_DCD  0x80

/* YAM control and status line mapping */
#define TXRDY    MSR_DCTS       /* transmitter ready status bit */
#define RXFLAG   MSR_RING       /* receiver FLAG status bit     */
#define RXCD     MSR_DCD        /* receiver CD status bit       */
#define FPGA_CFG MSR_DSR        /* FPGA configured status bit   */

#define PTT_ON   (MCR_RTS|MCR_OUT1|MCR_OUT2)    /* tx enable  */
#define PTT_OFF  (MCR_DTR|MCR_OUT1|MCR_OUT2)    /* tx disable */

/* BPQ */
#define DRIVER_INFO         1
#define ACCESS_TYPE         2
#define RELEASE_TYPE        3
#define SEND_PACKET         4
#define GET_ADDRESS         6
#define SET_RCV_MODE        20
#define SET_MULTICAST_LIST  22
#define GET_MULTICAST_LIST  23

#define BPQ_BUFF_SIZE      (328+16)             /* TF27b size + overhead  */
#define BPQ_RX_BUF_NR         8                 /* number of rx buffers  */
#define BPQ_TYPE_SIZE         2
#define BPQ_ETHADDR_SIZE      6

/* SCC */
#define SCCCHANS    4

#define PCLK        4915200L
#define TC75HZ      65534       /* 75 Hz zero count int */

/* KISS */
#define KISSPORTS   4

#define FEND        0xC0
#define FESC        0xDB
#define TFEND       0xDC
#define TFESC       0xDD

#define DATA        0x00
#define RMNC        0x20
#define SMACK       0x80

#define TPAR        1
#define PPAR        2
#define WPAR        3
#define TAPAR       4
#define DPAR        5

/* L1 */
#define MIN_LEN     8

#define STUFF       5
#define FLAG        6
#define ABRT        7
#define FREE        30

/* CRC */
#define CRC_RESET   0xFFFF
#define CRC_MASK    0x8408
#define CRC_CHECK   0xF0B8

/*** Z8530 ******************************************************************/

/* Write Registers */

#define RES_EX_INT  0x10        /* WR0 */
#define RES_TX_IP   0x28
#define RES_ERR     0x30
#define RES_TX_CRC  0x80
#define RES_EOM     0xC0

#define EX_IE       0x01        /* WR1 */
#define TX_IE       0x02
#define RX_IE       0x10

#define RX_DIS      0xC8        /* WR3 */
#define RX_EN       0xC9

#define SDLC        0x20        /* WR4 */

#define TX_DIS      0x61        /* WR5 */
#define TX_EN       0x69
#define RTS         0x02

#define SDLC_FLAG   0x7E        /* WR7 */

#define MINT_EN     0x0A        /* WR9 */
#define HW_RES      0xC0

#define NRZ         0x80        /* WR10 */
#define NRZI        0xA0
#define ABUNDER     0x04

#define TRxC_RTxC   0x20        /* WR11 */
#define DPLL_RTxC   0x64
#define DPLL_BRG    0x74
#define TRxOUT_BRG  0x02
#define TRxOUT_DPLL 0x03

#define DPLL_EN     0x23        /* WR14 */
#define DPLL_DIS    0x60
#define DPLL_SRCBRG 0x80
#define DPLL_NRZI   0xE0
#define BRG_EN      0x63

#define ZC_IE       0x02        /* WR15 */
#define DCD_IE      0x08
#define SYNC_IE     0x10
#define CTS_IE      0x20


/* Read Registers */

#define DCD         0x08        /* RR0 */
#define HUNT        0x10
#define CTS         0x20
#define TX_EOM      0x40
#define ABORT       0x80

#define RESIDUE     0x0E        /* RR1 */
#define RES8        0x06
#define RX_OVR      0x20
#define CRC_ERR     0x40
#define SDLC_EOF    0x80

/*** Macros *****************************************************************/

#define LOBYTE(val)         ((val) & 0xFF)
#define HIBYTE(val)         ((val) >> 8)

#define intr                _cdecl _interrupt _far

#define idisable()          { _asm  pushf \
                              _asm  cli  }
#define irestore()          { _asm  popf }

#define debug(major)        if (debugdev == major) \
                                outp (PPI, inp (PPI) & 0xFE ^ 2)

#define error(err, cond)    if (cond) errmsg (err)

#define set(var, bit)       var |=  (1 << bit)
#define res(var, bit)       var &= ~(1 << bit)

#define VIDEO               (*(bytefp) 0x449 == 7 \
                            ? (bytefp) 0xB000009F \
                            : (bytefp) 0xB800009F)

/* IRQ control */
#define irqvec(irq)         (irq + (irq < 8 ? IRQ0 : IRQ8-8))

#define maskreg(irq)        (irq < 8 ? PIC0+1 : PIC1+1)

#define irqmask(irq)        (1 << (irq & 7))

#define eoi(irq)            if (irq >= 8) \
                                outp (PIC1, EOI); \
                            outp (PIC0, EOI)

#define irq_prio(irq)       outp (PIC0, SETPRIO | (irq-1) & 7)

#define isat()              (*(bytefp) 0xF000FFFE == 0xFC)

/* Modem */
#define BIOSBASE            ((wordfp) (modem->rs232 ? 0x400L : 0x408L))

/* PAR */
#define LPTBASE             ((wordfp) 0x408L)

/* RTC */
#define WRRTC(reg, val)     outp ((outp (0x70, reg), 0x71), val)
#define RDRTC(reg)          inp ((outp (0x70, reg), 0x71))

/* SCC */
#define SCCDELAY            inp (0xE4)      /* non-existing port (2s) */

#define WR0(val)            outp ((SCCDELAY, scc->ctrl), val)
#define WR8(val)            WR (8, val)
#define WR(reg, val)        outp ((SCCDELAY, outp (scc->ctrl, reg), \
                                   SCCDELAY, scc->ctrl), val)

#define RR0                 inp ((SCCDELAY, scc->ctrl))
#define RR8                 inp ((SCCDELAY, scc->data))
#define RR(reg)             inp ((SCCDELAY, outp (scc->ctrl, reg), \
                                  SCCDELAY, scc->ctrl))

#define not0(par)           (par ? par : 1)

/* KISS and YAM */
#define COMBASE             ((wordfp) 0x400L)

/* KISS */
#define WRCOM(reg, val)     outp (kiss->base + reg, val)
#define RDCOM(reg)          inp (kiss->base + reg)

/* L2 */
#define rxput(port, val)    l1put ((port) << 8 | (val))
#define rxeof(port, ignore) l1put ((port) << 8 | 0x8000 | (ignore) << 1)
#define rxdiscard(port)     l1put ((port) << 8 | 0x8001)

#define txget(port)         l1get ((port) << 8)
#define txrewind(port)      l1get ((port) << 8 | 0x8000)
#define txdiscard(port)     l1get ((port) << 8 | 0x8001)

#define TXEOF               0x8000
#define TXLAST              0x8001

/*** Enum *******************************************************************/

enum {LOOP, MODEM, PAR, YAM, BPQ, SCC, KISS, RTC, NODEV};   /* Major Ports */

enum {L1NAME, L1TYPE, L1ADDR, L1IRQ, L1BAUD, L1ERR, L1CLR}; /* L1 Infos */

enum {DSCC, OSCC, USCC};                                    /* SCC Types */

enum {DISABLE, SOFTCLK, HARDCLK, DF9IC, PA0HZP, CLKCHAN};   /* Modem Types */

enum {IDLE, DWAIT, DELAY, ACTIVE, FLUSH, TAIL, TXCRC};      /* TX States */

enum {CRCOFF, CRCSMACK, CRCRMNC};                           /* CRC Modes */

/*** Types ******************************************************************/

typedef unsigned char       byte;
typedef unsigned short      word;
typedef unsigned long       dword;

typedef byte _far *         bytefp;
typedef word _far *         wordfp;

typedef void (intr *procfp) ();

/*** Struct *****************************************************************/

typedef struct buf_t
{
    struct buf_t *next;
    byte data[];
}
buf_t;

typedef struct modem_t
{
    byte l2port;

    byte port;
    bool rs232;
    word port_out, port_inp;
    byte ptt_on, ptt_off, data_out, data_inp;

    word baud, baud_delay, baud_cnt, send_cnt;

    byte dcd_cnt, bit_cnt, idle, shift, dpll, level;
    word frm_len, crc, fcs;
    bool dcd, sync;

    byte tx_state, tx_bitcnt, tx_idle;
    word tx_timer, tx_fcs, tx_shift;
    bool tx_kicked;
}
modem_t;

typedef struct par_t
{
    byte l2port;

    byte port;
    word port_out, port_inp;
    byte irq;
    procfp old_int;

    word baud;

    byte bit_cnt, idle, shift, lastbit;
    word frm_len, crc, fcs, dcd_cnt, dcd_timer;
    bool dcd, sync;

    byte tx_state, tx_bitcnt, tx_idle, tx_bit;
    word tx_timer, tx_fcs, tx_shift;
    bool tx_kicked;

    dword scr;
}
par_t;

typedef struct yam_t
{
    byte l2port;

    byte port;
    word port_io;
    byte irq;
    procfp old_int;

    word baud;

    byte tx_state, tx_timer, tx_fcsl, tx_fcsh;
    byte rx_fcsl, rx_fcsh, dcd, index;
    bool start_rx, tx_kicked;
}
yam_t;

typedef struct bpq_t
{
    byte l2port;

    byte int_nr;
    byte driv_class;
    word driv_type;
    byte driv_ifnum;
    int  handle;
    bool okay;

    byte e_type [BPQ_TYPE_SIZE];
    byte tx_buf [BPQ_BUFF_SIZE];
    byte ethaddr[BPQ_ETHADDR_SIZE];
    byte bpqaddr[BPQ_ETHADDR_SIZE];

    byte rx_buf [BPQ_RX_BUF_NR][BPQ_BUFF_SIZE];
    word rx_size[BPQ_RX_BUF_NR];
    bool rx_read_nr;
    bool rx_fill_nr;

    byte tx_state, tx_timer;
    word tx_count;
    bool tx_kicked;
}
bpq_t;

typedef struct scc_t
{
    byte l2port;

    byte chan;
    word data, ctrl;
    bool clkchan;

    word baud;
    byte modem;

    bool dcd, rx_err;
    word rx_len, dcd_timer;

    byte tx_state;
    word tx_timer;

    byte rr0;

    word errors;
}
scc_t;

typedef struct sccboard_t
{
    word base;
    byte irq;
    char modems[5];

    byte data[4];
    short ctrl;
}
sccboard_t;

typedef struct kiss_t
{
    byte l2port;

    byte port;
    word base;
    byte irq;
    procfp old_int;

    word baud;

    byte rx_state, tx_state;
    word rx_len, dcd_timer;
    bool tx_kicked;
    byte fesc;

    bool crc_ok;
    byte tx_mode, rx_mode;
    word tx_crc, rx_crc;

    word errors;

    buf_t *parlst;
    byte bufp;
}
kiss_t;

typedef struct port_t
{
    byte major;

    union
    {
        void    *minor;
        modem_t *modem;
        par_t   *par;
        yam_t   *yam;
        bpq_t   *bpq;
        scc_t   *scc;
        kiss_t  *kiss;
    };
}
port_t;

/*** Prototypes *************************************************************/

void l2rx (void);                             /* TF */
void l2tx (void);
void l2rest (void);
void lx (void);
void mainf (void);
bool l1put (word);
word l1get (word);
bool info_avl (void);
buf_t *allocb (void);
void dealoc (buf_t *);

void intr iface_mac (void);                /* IFACE */
word   _fastcall call_mac (word);
void   _cdecl chain (procfp);
byte   _cdecl random (void);
wordfp _cdecl set_stk (wordfp);
bool   _cdecl mtasker (void);
void   _cdecl bpq_rx_callback (void);

void bpq_rxtx (void);                       /* BPQ */

void int_tx (scc_t *);                      /* SCC */
void int_ex (scc_t *);
void int_rx (scc_t *);
void int_sp (scc_t *);

void set_pit(void);                         /* Misc */

/*** MAC Parameters *********************************************************/

byte defPp    = 32,
     defWp    = 10,
     defTp    = 25,
     defXp    = 1,
     defCp    = 0,
     defDp    = 0;

word defTAp   = 1;                          /* TX Tail */

/*** Default Parameters *****************************************************/

word defInt   = 0xFD,                       /* AX25_MAC interrupt */
     bufcnt   = MAC_BUFS;

bool logo     = true,                       /* Options */
     interlock= false;
     carrier  = false;

byte lastdev  = LOOP,                       /* Last attached device */
     debugdev = NODEV,                      /* Debug device */

     clockdev = LOOP,                       /* Clock device */
     clkprio  = 18,                         /* Ticks per second */
     clkstep  = 6;                          /* max. increment of ticks */

/* Ports */
byte l1ports;
port_t l1port[L1PORTS];

byte modems;
modem_t modemtab[MODEMS], *modemend = modemtab;
word pit0_const, int8_delay, tick_delay;

byte parports;
par_t partab;

byte yamports;
yam_t yamtab;

byte bpqports;
bpq_t bpqtab;
union  _REGS inregs;
union  _REGS outregs;
struct _SREGS segregs;
int    mac_cs;
int    mac_ds;

byte sccs, sccirq, scctype = DSCC;
scc_t scctab[SCCCHANS], *sccend = scctab, *sccclkchan = &scctab[0];

byte kissports;
kiss_t kisstab[KISSPORTS], *kissend = kisstab;

/*** Globals ****************************************************************/

extern byte Cpar[], Dpar[], Ppar[], Tpar[], Wpar[], Xpar[];
extern word TApar[], ticks;
extern void *lnktbl;

procfp old_int8, old_int1C, old_rtcint, old_sccint, old_defInt;

byte *bufbgn, *bufend;

word local_stack[STACK];
wordfp global_stack;
bool semaphore;

bool clock_error;
short clock_int8_count;

byte rxinfo, txinfo, attrib;
bool l2call;
word chbuf;

/*** IRQ Init ***************************************************************/

procfp irq_init (byte irq, procfp handler)
{
    procfp old_int;
    byte vec, pic;

    old_int = _dos_getvect (vec = irqvec (irq));
    _dos_setvect (vec, handler);
    outp (pic, inp (pic = maskreg (irq)) & ~irqmask (irq));

    return old_int;
}

/*** IRQ Reset **************************************************************/

void irq_reset (byte irq, procfp handler)
{
    byte pic;

    outp (pic, inp (pic = maskreg (irq)) | irqmask (irq));
    _dos_setvect (irqvec (irq), handler);
}

/*** Level 2 ****************************************************************/

void level_2 (void)
{
    _disable ();

    if (!semaphore)
    {
        semaphore = true;
        _enable ();
        global_stack = set_stk (local_stack + STACK);

        l2rx();
        l2tx();
        l2rest();

        set_stk (global_stack);
        semaphore = false;
    }
    else
        _enable ();
}

/*** Level X ****************************************************************/

void _cdecl level_x (void)
{
    _disable ();

    if (!semaphore)
    {
        semaphore = true;
        _enable ();
        global_stack = set_stk (local_stack + STACK);

        lx ();

        set_stk (global_stack);
        semaphore = false;
    }
    else
        _enable ();
}

/*** L1 Info ****************************************************************/

word l1info (byte port, byte info)
{
    static char *modems[] = {"SOFTCLK", "HARDCLK", "DF9IC", "PA0HZP"};
    modem_t *modem;
    kiss_t *kiss;
    scc_t *scc;
    char *name;

    switch (l1port[port].major)
    {
    case MODEM:
        modem = l1port[port].modem;

        switch (info)
        {
        case L1NAME: name    = modem->rs232 ? "COM1" : "LPT1";
                     name[3] = modem->port + '1';
                     return (word) name;

        case L1TYPE: return (word) "MODEM";
        case L1ADDR: return modem->port_out - (modem->rs232 ? 4 : 0);
        case L1BAUD: return modem->baud;
        }
        break;

    case PAR:
        switch (info)
        {
        case L1NAME: name    = "LPT1";
                     name[3] = partab.port + '1';
                     return (word) name;

        case L1TYPE: return (word) "PAR96";
        case L1ADDR: return partab.port_out;
        case L1IRQ:  return partab.irq;
        case L1BAUD: return partab.baud;
        }
        break;

    case YAM:
        switch (info)
        {
        case L1NAME: name    = "COM1";
                     name[3] = yamtab.port + '1';
                     return (word) name;

        case L1TYPE: return (word) "YAM96";
        case L1ADDR: return yamtab.port_io;
        case L1IRQ:  return yamtab.irq;
        case L1BAUD: return yamtab.baud;
        }
        break;

    case BPQ:
        switch (info)
        {
        case L1NAME: return "ETHER";

        case L1TYPE: return (word) "BPQ ";
        case L1ADDR: return bpqtab.int_nr;
        case L1IRQ:  return 0;
        case L1BAUD: return 0;
        }
        break;

    case SCC:
        scc = l1port[port].scc;

        switch (info)
        {
        case L1NAME: name    = "SCC0";
                     name[3] = scc->chan + '0';
                     return (word) name;

        case L1TYPE: return (word) modems[scc->modem - 1];
        case L1ADDR: return scctype == OSCC ? scc->ctrl : scc->data;
        case L1IRQ:  return sccirq;
        case L1BAUD: return scc->baud;

        case L1CLR:  scc->errors = 0;
        case L1ERR:  return scc->errors;
        }
        break;

    case KISS:
        kiss = l1port[port].kiss;

        switch (info)
        {
        case L1NAME: name    = "COM1+";
                     name[3] = kiss->port + '1';
                     name[4] = kiss->crc_ok ? '+' : '\0';
                     return (word) name;

        case L1TYPE: return (word) "KISS";
        case L1ADDR: return kiss->base;
        case L1IRQ:  return kiss->irq;
        case L1BAUD: return kiss->baud;

        case L1CLR:  kiss->errors = 0;
        case L1ERR:  return kiss->errors;
        }
    }

    return info == L1NAME ? (word) "NULL" : 0;
}

/*** L1 Parameter ***********************************************************/

void l1par (byte port, byte par, word val)
{
    buf_t *parbuf, *lastbuf;
    kiss_t *kiss;

    if (l1port[port].major == KISS)
    {
        kiss = l1port[port].kiss;

        parbuf = allocb ();
        parbuf->next = NULL;

        parbuf->data[0] = FEND;
        parbuf->data[1] = par;
        parbuf->data[2] = val <= 255 ? val : 255;
        parbuf->data[3] = FEND;

        idisable ();

        if (lastbuf = kiss->parlst)
        {
            while (lastbuf->next)
                lastbuf = lastbuf->next;

            lastbuf->next = parbuf;
        }
        else
        {
            kiss->parlst = parbuf;

            if (!kiss->tx_state)
            {
                WRCOM (0, parbuf->data[kiss->bufp++]);
                set (txinfo, kiss->l2port);
                kiss->tx_state = 6;
            }
        }

        irestore ();
    }
}

/*** Lock Random ************************************************************/

word _cdecl lock_random(void)
{
    /*
     * if we allow only one transmitter at a time then
     * we fake the random value to prevent other 
     * transmissions. Duplex transmissions will not be affected
     */
    if((interlock == true) && (txinfo != 0))
    {
        return 256; /* will defeat any Persist value... */
    }
    return (word) random();
}

/*** Is DCD? ****************************************************************/

bool iscd (byte port)
{
    int nr;

    if (!Dpar[port])
    {
        switch (l1port[port].major)
        {
        case MODEM: return l1port[port].modem->dcd;
        case PAR:   return partab.dcd;
        case YAM:   return yamtab.dcd;
        case BPQ:   return bpqtab.rx_size[bpqtab.rx_read_nr] != 0;
        case SCC:   return l1port[port].scc->dcd;
        }
    }

    return false;
}

/*** Kick TX ****************************************************************/

void kicktx (byte port)
{
    kiss_t *kiss;
    scc_t *scc;

    if (Xpar[port])
    {
        switch (l1port[port].major)
        {
        case MODEM:
            l1port[port].modem->tx_kicked = true;
            return;

        case PAR:
            idisable ();

            partab.tx_kicked = true;

            if (!partab.tx_state)
            {
                partab.tx_timer = Dpar[port] ? 1 : not0 (Wpar[port]);
                partab.tx_state = DWAIT;
            }

            irestore ();
            return;

        case YAM:
            idisable ();

            yamtab.tx_kicked = true;

            if (!yamtab.tx_state)
            {
                yamtab.tx_timer = Dpar[port] ? 1 : not0 (Wpar[port]);
                yamtab.tx_state = DWAIT;
            }

            irestore ();
            return;

        case BPQ:
            idisable ();

            bpqtab.tx_kicked = true;

            if (!bpqtab.tx_state)
            {
                bpqtab.tx_state = ACTIVE;
            }

            irestore ();
            return;

        case SCC:
            scc = l1port[port].scc;

            idisable ();

            switch (scc->tx_state)
            {
            case IDLE:
                scc->tx_timer = Dpar[port] ? 1 : not0 (Wpar[port]);
                scc->tx_state = DWAIT;
                break;

            case TAIL:
                scc->tx_timer = 1;
                scc->tx_state = DELAY;
            }

            irestore ();
            return;

        case KISS:
            kiss = l1port[port].kiss;

            idisable ();

            if (kiss->tx_state)
                kiss->tx_kicked = true;
            else
            {
                WRCOM (0, FEND);
                set (txinfo, kiss->l2port);
                kiss->tx_state = 1;
            }

            irestore ();
            return;
        }
    }

    txdiscard (port);
}

/*** SCC Clock **************************************************************/

void scc_clk (scc_t *scc, bool tx_on)
{
    word tconst;

    if (scc->modem == DF9IC)
    {
        WR (11, TRxC_RTxC);
        WR (14, DPLL_DIS);
    }
    else
    {
        tconst = PCLK / 64L / scc->baud;
        if (tx_on)
            tconst *= 32;
        tconst -= 2;

        WR (12, LOBYTE (tconst));
        WR (13, HIBYTE (tconst));

        WR (11, tx_on ? DPLL_BRG | TRxOUT_BRG
                      : (scc->modem == PA0HZP ? DPLL_RTxC | TRxOUT_BRG
                                              : DPLL_RTxC | TRxOUT_DPLL));

        WR (14, DPLL_SRCBRG);
        WR (14, DPLL_NRZI);
        WR (14, DPLL_EN);
    }
}

/*** SCC TX *****************************************************************/

void scc_tx (scc_t *scc, bool tx_on)
{
    if (scc->modem == SOFTCLK)
    {
        WR (3, RX_DIS);
        WR (5, TX_DIS);

        scc_clk (scc, tx_on);

        if (tx_on)
            WR (5, TX_EN | RTS);
        else
            WR (3, RX_EN);
    }
    else
        WR (5, tx_on ? TX_EN | RTS : TX_EN);
}

/*** Timer ******************************************************************/

void timer (word ticks)
{
    static word serv_cnt = 10;
    bytefp video = VIDEO;
    modem_t *modem;
    kiss_t *kiss;
    scc_t *scc;
    byte port;

    for (modem = modemtab; modem < modemend; modem++)
    {
        _disable ();

        if (modem->tx_timer)
            if (modem->tx_timer <= ticks)
                modem->tx_timer = 0;
            else
                modem->tx_timer -= ticks;

        _enable ();
    }

    if (parports)
    {
        _disable ();

        if (partab.tx_timer)
        {
            if (partab.tx_timer <= ticks)
            {
                partab.tx_timer = 0;

                if (partab.tx_state == DWAIT)
                {
                    if (   Dpar[partab.l2port] || !partab.dcd
                           && lock_random () <= Ppar[partab.l2port])
                    {
                        set (txinfo, partab.l2port);
                        partab.tx_timer = Tpar[partab.l2port];
                        partab.tx_state = DELAY;
                    }
                    else
                    {
                        partab.tx_timer = not0 (Wpar[partab.l2port]);
                    }
                }
            }
            else
            {
                partab.tx_timer -= ticks;
            }
        }

        if (partab.dcd_timer)
        {
            if (partab.dcd_timer <= ticks)
            {
                partab.dcd_timer = 0;

                set (rxinfo, partab.l2port);
            }
            else
            {
                partab.dcd_timer -= ticks;
            }
        }

        _enable ();
    }

    if (yamports)
    {
        _disable ();

        if (yamtab.tx_timer)
            if (yamtab.tx_timer <= ticks)
            {
                yamtab.tx_timer = 0;

                if (yamtab.tx_state == DWAIT)
                    if (   Dpar[yamtab.l2port] || !yamtab.dcd
                           && lock_random () <= Ppar[partab.l2port])
                    {
                        set (txinfo, yamtab.l2port);
                        yamtab.tx_timer = Tpar[yamtab.l2port];
                        yamtab.tx_state = DELAY;
                        /*  switch the PTT */
                        outp (yamtab.port_io + MCR, PTT_ON);
                        /*  enable modem status interruprs */
                        inp  (yamtab.port_io + MSR);
                        outp (yamtab.port_io + IER, IER_MSR);
                    }
                    else
                        yamtab.tx_timer = not0 (Wpar[yamtab.l2port]);
            }
            else
                yamtab.tx_timer -= ticks;

        _enable ();
    }

    if (bpqports)
    {
        _disable ();

        if((bpqtab.tx_state) || (bpqtab.rx_size[bpqtab.rx_read_nr]))
                bpq_rxtx();

        _enable ();
    }

    for (scc = scctab; scc < sccend; scc++)
    {
        if (scc->modem)
        {
            _disable ();

            if (scc->tx_timer)
                if (scc->tx_timer <= ticks)
                {
                    scc->tx_timer = 0;

                    switch (scc->tx_state)
                    {
                    case DWAIT:
                        if (Dpar[scc->l2port] || !scc->dcd
                            && lock_random () <= Ppar[scc->l2port])
                        {
                            scc_tx (scc, true);
                            set (txinfo, scc->l2port);
                            scc->tx_timer = not0 (Tpar[scc->l2port]);
                            scc->tx_state = DELAY;
                        }
                        else
                            scc->tx_timer = not0 (Wpar[scc->l2port]);
                        break;

                    case DELAY:
                        WR0 (RES_TX_CRC);
                        WR8 (txget (scc->l2port));
                        WR0 (RES_EOM);
                        scc->tx_state = ACTIVE;
                        break;

                    case TAIL:
                        scc_tx (scc, false);
                        res (txinfo, scc->l2port);
                        scc->tx_state = IDLE;
                    }
                }
                else
                    scc->tx_timer -= ticks;

            if (scc->dcd_timer)
                if (scc->dcd_timer <= ticks)
                {
                    scc->dcd_timer = 0;

                    set (rxinfo, scc->l2port);
                }
                else
                    scc->dcd_timer -= ticks;

            _enable ();
        }
    }

    for (kiss = kisstab; kiss < kissend; kiss++)
    {
        _disable ();

        if (kiss->dcd_timer)
        {
            if (kiss->dcd_timer <= ticks)
            {
                kiss->dcd_timer = 0;

                res (rxinfo, kiss->l2port);
            }
            else
            {
                kiss->dcd_timer -= ticks;
            }
        }

        _enable ();
    }

    if (serv_cnt <= ticks || l2call)
    {
        serv_cnt = 10;              /* 10/s */
        l2call = false;
        level_2 ();
    }
    else
        serv_cnt -= ticks;

    if (carrier && l1ports)
    {
        for (port = 1 << l1ports-1; port; port >>= 1, video--)
        {
            if (attrib)
                *video = attrib;
            *--video = txinfo & port ? 'S' : (rxinfo & port ? 'R' : '');
        }
    }
}

/*** SCC Interrupt **********************************************************/

void (*handler[]) (scc_t *) = {int_tx, int_ex, int_rx, int_sp};

void intr scc_int (void)
{
    static word oldticks;
    word newticks;
    scc_t *scc;
    byte rr2;

    debug (SCC);

    _disable ();
    eoi (sccirq);

    for (scc = scctab; scc < sccend;)
        if (RR (3))
        {
            scc++;
            rr2 = RR (2);
            handler[rr2 >> 1 & 3] (&scctab[rr2 >> 3 ^ 1]);
            scc = scctab;
        }
        else
            scc += 2;

    if (clockdev == SCC && (newticks = ticks - oldticks))
    {
        oldticks = ticks;
        _enable ();
        timer (newticks);
    }
}

/*** TX Handler *************************************************************/

void int_tx (scc_t *scc)
{
    word val;

    switch (scc->tx_state)
    {
    case ACTIVE:
        if (RR0 & TX_EOM)
        {
            scc->errors++;

            txrewind (scc->l2port);
            scc->tx_state = FLUSH;
        }
        else
            if ((val = txget (scc->l2port)) & TXEOF)
            {
                WR (10, scc->modem == DF9IC ? NRZ : NRZI);
                WR0 (RES_TX_IP);
                scc->tx_state = FLUSH;
            }
            else
                WR8 (val);
        break;

    case FLUSH:
        WR (10, scc->modem == DF9IC ? NRZ | ABUNDER : NRZI | ABUNDER);

        if ((val = txget (scc->l2port)) < TXEOF)
        {
            WR0 (RES_TX_CRC);
            WR8 (val);
            WR0 (RES_EOM);
            scc->tx_state = ACTIVE;
        }
        else
        {
            WR0 (RES_TX_IP);
            scc->tx_timer = TApar[scc->l2port] + clkstep;
            scc->tx_state = TAIL;
        }
    }
}

/*** Extern Status Change ***************************************************/

void int_ex (scc_t *scc)
{
    static byte cnt;
    byte rr0, delta;

    delta = (rr0 = RR0) ^ scc->rr0;

    if (scc->modem)
    {
        if (scc->rx_len && rr0 & (ABORT | HUNT))
        {
            rxdiscard (scc->l2port);
            scc->rx_len = 0;
            scc->rx_err = false;
        }

        if (delta & (Cpar[scc->l2port] ? HUNT : DCD))
        {
            if (scc->dcd = Cpar[scc->l2port] ? !(rr0 & HUNT) : rr0 & DCD)
            {
                if (!(scc->dcd_timer = Cpar[scc->l2port]))
                    set (rxinfo, scc->l2port);
            }
            else
            {
                res (rxinfo, scc->l2port);
                scc->dcd_timer = 0;
            }
        }
    }

    if (scc->clkchan && (scctype == OSCC || delta & CTS))
        ticks += cnt++ == 2 ? cnt = 0, 2 : 1;

    scc->rr0 = rr0;
    WR0 (RES_EX_INT);
}

/*** RX Handler *************************************************************/

void int_rx (scc_t *scc)
{
    byte rr8;

    rr8 = RR8;

    if (!scc->rx_err)
    {
        scc->rx_err = rxput (scc->l2port, rr8);
        scc->rx_len++;
    }
}

/*** Special RX Condition ***************************************************/

void int_sp (scc_t *scc)
{
    byte rr1;

    rr1 = RR (1);
    RR8;

    if (rr1 & RX_OVR)
    {
        scc->errors++;
        scc->rx_err = true;
    }

    if (rr1 & SDLC_EOF)
    {
        if (   !scc->rx_err
            && scc->rx_len >= MIN_LEN+1
            && (rr1 & (CRC_ERR | RESIDUE)) == RES8)
        {
            rxeof (scc->l2port, 1);
            l2call = true;
        }
        else
            if (scc->rx_len)
                rxdiscard (scc->l2port);

        scc->rx_len = 0;
        scc->rx_err = false;
    }

    WR0 (RES_ERR);
}

/*** SCC Hardware Reset *****************************************************/

void scc_hw_reset (void)
{
    scc_t *scc;
    word delay;

    for (scc = scctab; scc < &scctab[SCCCHANS]; scc += 2)
    {
        RR0;
        WR (9, HW_RES);
        for (delay = 100; delay--;);
    }
}

/*** SCC Init ***************************************************************/

void scc_init (void)
{
    scc_t *scc;

    if (clockdev == SCC)
        sccclkchan->clkchan = true;

    scc_hw_reset ();

    old_sccint = irq_init (sccirq, scc_int);

    for (scc = scctab; scc < sccend; scc++)
    {
        if (scc->modem)
        {
            if (scc->baud == 300)
                TApar[scc->l2port] = 4;

            WR (4, SDLC);
            WR (10, scc->modem == DF9IC ? NRZ | ABUNDER : NRZI | ABUNDER);
            WR (7, SDLC_FLAG);
            WR (3, RX_DIS);
            WR (5, TX_DIS);

            scc_clk (scc, false);

            WR (3, RX_EN);
            WR (5, scc->modem == SOFTCLK ? TX_DIS : TX_EN);

            WR (15, scc->clkchan ? SYNC_IE | DCD_IE | CTS_IE
                                 : SYNC_IE | DCD_IE);
            WR0 (RES_EX_INT);
            WR0 (RES_EX_INT);
            WR (1, EX_IE | TX_IE | RX_IE);
        }
        else
            if (scc->clkchan)
            {
                if (scctype == OSCC)
                {
                    WR (12, LOBYTE (TC75HZ));
                    WR (13, HIBYTE (TC75HZ));
                    WR (14, BRG_EN);

                    WR (15, ZC_IE);
                }
                else
                    WR (15, CTS_IE);

                WR0 (RES_EX_INT);
                WR0 (RES_EX_INT);
                WR (1, EX_IE);
            }

        scc->rr0 = RR0;
    }

    for (scc = scctab; scc < sccend; scc += 2)
    {
        WR (2, scc->chan / 2 * 0x10);
        WR (9, MINT_EN);
    }
}

/*** SCC Reset **************************************************************/

void scc_reset (void)
{
    irq_reset (sccirq, old_sccint);

    scc_hw_reset ();
}

/*** SCC Detect *************************************************************/

bool scc_detect (void)
{
    scc_t *scc;

    for (scc = scctab; scc < sccend; scc++)
    {
        RR0;
        if (   (WR (13, 0x55), RR (13) != 0x55)
            || (WR (13, 0xAA), RR (13) != 0xAA))
            return false;
    }

    return true;
}

/*** Modem RX ***************************************************************/

void modem_rx (modem_t *modem)
{
    if ((inp (modem->port_inp) ^ modem->level) & modem->data_inp)
    {
        modem->level ^= 0xFF;

        if (!Cpar[modem->l2port] || modem->dpll == 2)
        {
            if (modem->dcd_cnt < Cpar[modem->l2port])
                modem->dcd_cnt++;
            else
            {
                modem->dcd = true;
                set (rxinfo, modem->l2port);
            }
        }
        else
        {
            if (modem->dcd_cnt)
                modem->dcd_cnt--;
            else
            {
                modem->dcd = false;
                res (rxinfo, modem->l2port);
            }
        }

        modem->dpll = 0;

        if (modem->idle < STUFF)
        {
            modem->idle = 0;
            goto shift_bit;
        }

        if (modem->idle == FLAG)
        {
            if (modem->frm_len)
            {
                if (   modem->sync
                    && modem->frm_len >= MIN_LEN+2
                    && modem->bit_cnt == 1
                    && modem->fcs     == CRC_CHECK)
                {
                    rxeof (modem->l2port, 2);
                    l2call = true;
                }
                else
                    rxdiscard (modem->l2port);

                modem->frm_len = 0;
            }

            modem->bit_cnt = 8;
            modem->sync    = true;
            modem->crc     = CRC_RESET;
        }

        modem->idle = 0;
    }
    else
    {
        if (++modem->dpll == 4)
        {
            modem->dpll = 1;

            if (modem->idle < FREE)
            {
                if (++modem->idle < ABRT)
                {
shift_bit:          if (modem->sync)
                    {
                        if (modem->shift >>= 1, modem->dpll)
                            modem->shift |= 0x80;

                        modem->crc = (modem->crc >> 1) ^ ((modem->dpll ^
                                      modem->crc) & 1 ? CRC_MASK : 0);

                        if (--modem->bit_cnt == 0)
                        {
                            modem->sync    = !rxput (modem->l2port,
                                                     modem->shift);
                            modem->fcs     = modem->crc;
                            modem->bit_cnt = 8;
                            modem->frm_len++;
                        }
                    }
                }
                else
                    modem->sync = false;
            }
            else
            {
                modem->dcd = false;
                res (rxinfo, modem->l2port);
            }
        }
    }
}

/*** Modem TX ***************************************************************/

void modem_tx (modem_t *modem)
{
    if (modem->tx_state || modem->tx_kicked)
    {
        if (modem->tx_idle == STUFF && (modem->tx_state == ACTIVE
                                    ||  modem->tx_state == FLUSH))
        {
            outp (modem->port_out, inp (modem->port_out) ^ modem->data_out);
            modem->tx_idle = 0;
        }
        else
        {
            if (!modem->tx_bitcnt || modem->tx_state <= DWAIT)
            {
                switch (modem->tx_state)
                {
                case IDLE:
                    if (modem->dcd && !Dpar[modem->l2port])
                        return;

                    modem->tx_timer = Wpar[modem->l2port];
                    modem->tx_state = DWAIT;

                case DWAIT:
                    if (!Dpar[modem->l2port])
                    {
                        if (modem->tx_timer)
                            return;

                        if (modem->dcd || Ppar[modem->l2port] < lock_random ())
                        {
                            modem->tx_state = IDLE;
                            return;
                        }
                    }

                    outp (modem->port_out, modem->ptt_on);
                    set (txinfo, modem->l2port);
                    modem->tx_timer = Tpar[modem->l2port];
                    modem->tx_state = DELAY;

                case DELAY:
                    if (modem->tx_timer)
                    {
                        modem->tx_shift  = SDLC_FLAG;
                        modem->tx_bitcnt = 8;
                        break;
                    }

                next_frm:
                    modem->tx_fcs   = CRC_RESET;
                    modem->tx_idle  = 0;
                    modem->tx_state = ACTIVE;

                case ACTIVE:
                    if ((modem->tx_shift = txget (modem->l2port)) < TXEOF)
                    {
                        modem->tx_bitcnt = 8;
                        break;
                    }

                    modem->tx_kicked = modem->tx_shift != TXLAST;
                    modem->tx_state  = FLUSH;
                    modem->tx_shift  = ~modem->tx_fcs;
                    modem->tx_bitcnt = 16;
                    break;

                case FLUSH:
                    modem->tx_shift  = SDLC_FLAG;
                    modem->tx_bitcnt = 8;
                    modem->tx_timer  = TApar[modem->l2port] + clkstep;
                    modem->tx_state  = TAIL;
                    break;

                case TAIL:
                    if (modem->tx_kicked)
                        goto next_frm;

                    if (modem->tx_timer)
                    {
                        modem->tx_shift  = SDLC_FLAG;
                        modem->tx_bitcnt = 8;
                        break;
                    }

                    outp (modem->port_out, modem->ptt_off);
                    res (txinfo, modem->l2port);
                    modem->tx_state = IDLE;
                    return;
                }
            }

            if (modem->tx_shift & 1)
                modem->tx_idle++;
            else
            {
                outp (modem->port_out,
                      inp (modem->port_out) ^ modem->data_out);
                modem->tx_idle = 0;
            }

            modem->tx_fcs = (modem->tx_fcs >> 1) ^ ((modem->tx_shift ^
                             modem->tx_fcs) & 1 ? CRC_MASK : 0);

            modem->tx_shift >>= 1;
            modem->tx_bitcnt--;
        }
    }
}

/*** Modem Init *************************************************************/

void modem_init (void)
{
    modem_t *modem;

    for (modem = modemtab; modem < modemend; modem++)
    {
        if (modem->baud == 300)
            TApar[modem->l2port] = 4;

        outp (modem->port_out, modem->ptt_off);
        if (modem->rs232)
            outp (modem->port_out - 1, 0x43);   /* BREAK (TxD +12V) */
    }

    set_pit();
}

void set_pit(void)
{
    outp (PIT, LOBYTE (pit0_const));            /* set time constant */
    outp (PIT, HIBYTE (pit0_const));
    clock_error = false;
}

/*** Modem Reset ************************************************************/

void modem_reset (void)
{
    modem_t *modem;

    outp (PIT, 0);                              /* reset timer */
    outp (PIT, 0);

    for (modem = modemtab; modem < modemend; modem++)
    {
        outp (modem->port_out, 0);
        if (modem->rs232)
            outp (modem->port_out - 1, 3);      /* reset TxD */
    }
}

/*** RTC Interrupt **********************************************************/

void intr rtc_int (void)
{
    static byte step[] = {2, 1, 2, 1, 2, 1, 2, 2, 1, 2, 1, 2, 1, 2, 1, 2};
    static byte cnt;
    word newticks;

    eoi (8);

    if ((RDRTC (12) & 0xC0) == 0xC0)
    {
        if (clockdev == RTC)
        {
            ticks += newticks = step[cnt++ % 16];

            _enable ();
            timer (newticks);
        }
        else
        {
            /* check int8 interrupt speed */

            /*
             * Keep timer at high speed - PE1DNN
             *
             * The counter is incremented by the INT8 ISR.
             *
             * The lowest baudrate is 300 baud which needs a 900 Hz clock on
             * INT8. Every RTC interrupt 900/64 = ~14 int8 interrupts shall
             * be counted. We check for 10, if lower we have to re-program
             * the timer. Unfortunately this only works on PC-AT or better...
             */
            if(clock_int8_count < 10)
            {
                /*
                 * Since it is not always successful due to race conditions
                 * we also raise a flag. On the first application call the
                 * PIT will be initialized again
                 */
                set_pit();
                clock_error = true;
            }

            clock_int8_count = 0;
        }
    }
}

/*** RTC Init ***************************************************************/

void rtc_init (void)
{
    old_rtcint = irq_init (8, rtc_int);

    WRRTC (10, 0x2A);
    WRRTC (11, 0x42);
}

/*** RTC Reset **************************************************************/

void rtc_reset (void)
{
    WRRTC (11, 0x02);
    WRRTC (10, 0x26);

    irq_reset (8, old_rtcint);
}

/*** PAR RX *****************************************************************/

void par_rx (void)
{
    byte buf, bit, cnt = 16;

    while (cnt--)
    {
        bit = (buf = inp (partab.port_inp)) & RXD;
        outp (partab.port_out, PTT);
        outp (partab.port_out, PTT | BURST);

        partab.scr = (partab.scr << 1) | (bit != partab.lastbit);
        partab.lastbit = bit;

        bit =   !((word) partab.scr & 1)
              ^ !((word) partab.scr & 0x1000)
              ^ !(       partab.scr & 0x20000L);

        if (!bit)
        {
            if (partab.idle < STUFF)
            {
                partab.idle = 0;
                goto shift_bit;
            }

            if (partab.idle == FLAG)
            {
                if (   Cpar[partab.l2port]
                    && partab.dcd_cnt < DCDLIMIT
                    && ++partab.dcd_cnt == DCDCOUNT)
                {
                    partab.dcd = true;
                    partab.dcd_timer = Cpar[partab.l2port];
                }

                if (partab.frm_len)
                {
                    if (   partab.sync
                        && partab.frm_len >= MIN_LEN+2
                        && partab.bit_cnt == 1
                        && partab.fcs     == CRC_CHECK)
                    {
                        rxeof (partab.l2port, 2);
                        l2call = true;
                    }
                    else
                        rxdiscard (partab.l2port);

                    partab.frm_len = 0;
                }

                partab.bit_cnt = 8;
                partab.sync    = true;
                partab.crc     = CRC_RESET;
            }

            partab.idle = 0;
        }
        else
        {
            if (partab.idle < FREE)
            {
                if (++partab.idle < ABRT)
                {
shift_bit:          if (partab.sync)
                    {
                        if (partab.shift >>= 1, bit)
                            partab.shift |= 0x80;

                        partab.crc = (partab.crc >> 1) ^ ((bit ^
                                      partab.crc) & 1 ? CRC_MASK : 0);

                        if (--partab.bit_cnt == 0)
                        {
                            partab.sync    = !rxput (partab.l2port,
                                                     partab.shift);
                            partab.fcs     = partab.crc;
                            partab.bit_cnt = 8;
                            partab.frm_len++;
                        }
                    }
                }
                else
                {
                    if (Cpar[partab.l2port])
                    {
                        if (partab.dcd_cnt > DCDABORT)
                            partab.dcd_cnt -= DCDABORT;
                        else
                            partab.dcd_cnt = 0;

                        if (partab.dcd_cnt < DCDCOUNT)
                        {
                            partab.dcd = false;
                            partab.dcd_timer = 0;
                            res (rxinfo, partab.l2port);
                        }
                    }

                    partab.sync = false;
                }
            }
        }
    }

    if (!Cpar[partab.l2port])
        if (partab.dcd = buf & DCDBIT)
            set (rxinfo, partab.l2port);
        else
            res (rxinfo, partab.l2port);
}

/*** PAR TX *****************************************************************/

void par_tx (void)
{
    byte bit, cnt = 16;

    while (cnt--)
    {
        if (   partab.tx_idle == STUFF
            && (   partab.tx_state == ACTIVE
                || partab.tx_state == FLUSH))
        {
            partab.tx_bit ^= 1;
            partab.tx_idle = 0;
        }
        else
        {
            if (!partab.tx_bitcnt)
            {
                switch (partab.tx_state)
                {
                case DELAY:
                    if (partab.tx_timer)
                    {
                        partab.tx_shift  = SDLC_FLAG;
                        partab.tx_bitcnt = 8;
                        break;
                    }

                next_frm:
                    partab.tx_fcs   = CRC_RESET;
                    partab.tx_idle  = 0;
                    partab.tx_state = ACTIVE;

                case ACTIVE:
                    if ((partab.tx_shift = txget (partab.l2port)) < TXEOF)
                    {
                        partab.tx_bitcnt = 8;
                        break;
                    }

                    partab.tx_kicked = partab.tx_shift != TXLAST;
                    partab.tx_state  = FLUSH;
                    partab.tx_shift  = ~partab.tx_fcs;
                    partab.tx_bitcnt = 16;
                    break;

                case FLUSH:
                    partab.tx_shift  = SDLC_FLAG;
                    partab.tx_bitcnt = 8;
                    partab.tx_timer  = TApar[partab.l2port] + clkstep;
                    partab.tx_state  = TAIL;
                    break;

                case TAIL:
                    if (partab.tx_kicked)
                        goto next_frm;

                    if (partab.tx_timer)
                    {
                        partab.tx_shift  = SDLC_FLAG;
                        partab.tx_bitcnt = 8;
                        break;
                    }

                    outp (partab.port_out, PTT | BURST);
                    res (txinfo, partab.l2port);
                    partab.tx_state = IDLE;
                    return;
                }
            }

            if (partab.tx_shift & 1)
                partab.tx_idle++;
            else
            {
                partab.tx_bit ^= 1;
                partab.tx_idle = 0;
            }

            partab.tx_fcs = (partab.tx_fcs >> 1) ^ ((partab.tx_shift ^
                             partab.tx_fcs) & 1 ? CRC_MASK : 0);

            partab.tx_shift >>= 1;
            partab.tx_bitcnt--;
        }

        partab.scr =   (partab.scr << 1)
                     | (bit =    partab.tx_bit
                              ^ !((word) partab.scr & 0x800)
                              ^ !(       partab.scr & 0x10000L));

        outp (partab.port_out, bit);
        outp (partab.port_out, bit | BURST);
    }
}

/*** PAR Interrupt **********************************************************/

void intr par_int (void)
{
    debug (PAR);

    if (partab.tx_state > DWAIT)
        par_tx ();
    else
        par_rx ();

    eoi (partab.irq);
}

/*** PAR Init ***************************************************************/

void par_init (void)
{
    partab.old_int = irq_init (partab.irq, par_int);

    outp (partab.port_out, PTT | BURST);
    outp (partab.port_out + 2, inp (partab.port_out + 2) | IRQEN);
}

/*** PAR Reset **************************************************************/

void par_reset (void)
{
    outp (partab.port_out + 2, inp (partab.port_out + 2) & ~IRQEN);
    outp (partab.port_out, PTT | BURST);

    irq_reset (partab.irq, partab.old_int);
}

/*** YAM CRC tables *********************************************************/

char chktabl[256]=
{ 0x00,0x89,0x12,0x9b,0x24,0xad,0x36,0xbf,0x48,0xc1,0x5a,0xd3,0x6c,0xe5,0x7e,
  0xf7,0x81,0x08,0x93,0x1a,0xa5,0x2c,0xb7,0x3e,0xc9,0x40,0xdb,0x52,0xed,0x64,
  0xff,0x76,0x02,0x8b,0x10,0x99,0x26,0xaf,0x34,0xbd,0x4a,0xc3,0x58,0xd1,0x6e,
  0xe7,0x7c,0xf5,0x83,0x0a,0x91,0x18,0xa7,0x2e,0xb5,0x3c,0xcb,0x42,0xd9,0x50,
  0xef,0x66,0xfd,0x74,0x04,0x8d,0x16,0x9f,0x20,0xa9,0x32,0xbb,0x4c,0xc5,0x5e,
  0xd7,0x68,0xe1,0x7a,0xf3,0x85,0x0c,0x97,0x1e,0xa1,0x28,0xb3,0x3a,0xcd,0x44,
  0xdf,0x56,0xe9,0x60,0xfb,0x72,0x06,0x8f,0x14,0x9d,0x22,0xab,0x30,0xb9,0x4e,
  0xc7,0x5c,0xd5,0x6a,0xe3,0x78,0xf1,0x87,0x0e,0x95,0x1c,0xa3,0x2a,0xb1,0x38,
  0xcf,0x46,0xdd,0x54,0xeb,0x62,0xf9,0x70,0x08,0x81,0x1a,0x93,0x2c,0xa5,0x3e,
  0xb7,0x40,0xc9,0x52,0xdb,0x64,0xed,0x76,0xff,0x89,0x00,0x9b,0x12,0xad,0x24,
  0xbf,0x36,0xc1,0x48,0xd3,0x5a,0xe5,0x6c,0xf7,0x7e,0x0a,0x83,0x18,0x91,0x2e,
  0xa7,0x3c,0xb5,0x42,0xcb,0x50,0xd9,0x66,0xef,0x74,0xfd,0x8b,0x02,0x99,0x10,
  0xaf,0x26,0xbd,0x34,0xc3,0x4a,0xd1,0x58,0xe7,0x6e,0xf5,0x7c,0x0c,0x85,0x1e,
  0x97,0x28,0xa1,0x3a,0xb3,0x44,0xcd,0x56,0xdf,0x60,0xe9,0x72,0xfb,0x8d,0x04,
  0x9f,0x16,0xa9,0x20,0xbb,0x32,0xc5,0x4c,0xd7,0x5e,0xe1,0x68,0xf3,0x7a,0x0e,
  0x87,0x1c,0x95,0x2a,0xa3,0x38,0xb1,0x46,0xcf,0x54,0xdd,0x62,0xeb,0x70,0xf9,
  0x8f,0x06,0x9d,0x14,0xab,0x22,0xb9,0x30,0xc7,0x4e,0xd5,0x5c,0xe3,0x6a,0xf1,
  0x78
};
char chktabh[256]=
{ 0x00,0x11,0x23,0x32,0x46,0x57,0x65,0x74,0x8c,0x9d,0xaf,0xbe,0xca,0xdb,0xe9,
  0xf8,0x10,0x01,0x33,0x22,0x56,0x47,0x75,0x64,0x9c,0x8d,0xbf,0xae,0xda,0xcb,
  0xf9,0xe8,0x21,0x30,0x02,0x13,0x67,0x76,0x44,0x55,0xad,0xbc,0x8e,0x9f,0xeb,
  0xfa,0xc8,0xd9,0x31,0x20,0x12,0x03,0x77,0x66,0x54,0x45,0xbd,0xac,0x9e,0x8f,
  0xfb,0xea,0xd8,0xc9,0x42,0x53,0x61,0x70,0x04,0x15,0x27,0x36,0xce,0xdf,0xed,
  0xfc,0x88,0x99,0xab,0xba,0x52,0x43,0x71,0x60,0x14,0x05,0x37,0x26,0xde,0xcf,
  0xfd,0xec,0x98,0x89,0xbb,0xaa,0x63,0x72,0x40,0x51,0x25,0x34,0x06,0x17,0xef,
  0xfe,0xcc,0xdd,0xa9,0xb8,0x8a,0x9b,0x73,0x62,0x50,0x41,0x35,0x24,0x16,0x07,
  0xff,0xee,0xdc,0xcd,0xb9,0xa8,0x9a,0x8b,0x84,0x95,0xa7,0xb6,0xc2,0xd3,0xe1,
  0xf0,0x08,0x19,0x2b,0x3a,0x4e,0x5f,0x6d,0x7c,0x94,0x85,0xb7,0xa6,0xd2,0xc3,
  0xf1,0xe0,0x18,0x09,0x3b,0x2a,0x5e,0x4f,0x7d,0x6c,0xa5,0xb4,0x86,0x97,0xe3,
  0xf2,0xc0,0xd1,0x29,0x38,0x0a,0x1b,0x6f,0x7e,0x4c,0x5d,0xb5,0xa4,0x96,0x87,
  0xf3,0xe2,0xd0,0xc1,0x39,0x28,0x1a,0x0b,0x7f,0x6e,0x5c,0x4d,0xc6,0xd7,0xe5,
  0xf4,0x80,0x91,0xa3,0xb2,0x4a,0x5b,0x69,0x78,0x0c,0x1d,0x2f,0x3e,0xd6,0xc7,
  0xf5,0xe4,0x90,0x81,0xb3,0xa2,0x5a,0x4b,0x79,0x68,0x1c,0x0d,0x3f,0x2e,0xe7,
  0xf6,0xc4,0xd5,0xa1,0xb0,0x82,0x93,0x6b,0x7a,0x48,0x59,0x2d,0x3c,0x0e,0x1f,
  0xf7,0xe6,0xd4,0xc5,0xb1,0xa0,0x92,0x83,0x7b,0x6a,0x58,0x49,0x3d,0x2c,0x1e,
  0x0f
};

/*** YAM RX *****************************************************************/

void yam_rx (byte msr, byte rxb)
{
    byte temp;

    if (!(yamtab.dcd = msr & RXCD))          /*is carrier active ? */
    {
        res (rxinfo, yamtab.l2port);         /* no */
        yamtab.start_rx = 0;
    }
    else                                     /* carrier detected */
    {
        if (msr & RXFLAG)                    /* flag or data received ?     */
        {
            /* check if rx frame is pending and whether it's o.k. */
            if((yamtab.start_rx) && ((yamtab.rx_fcsl & yamtab.rx_fcsh) == 0xff))
            {
                rxeof (yamtab.l2port, 2);   /* FCS okay */
                l2call = true;
            }
            else
                rxdiscard (yamtab.l2port);  /* FCS error */

            yamtab.start_rx = 1;            /* ready to accept a new frame */
            yamtab.rx_fcsl  = 0x21;         /* initialize CRC generator    */
            yamtab.rx_fcsh  = 0xf3;
        }
        else                                /* packet data received        */
        {
            /* did we received a Flag after DCD high? */
            if(yamtab.start_rx)
            {
                /* Yes */
                rxput(yamtab.l2port, rxb);
                temp           = yamtab.rx_fcsl;
                yamtab.rx_fcsl = chktabl[temp] ^ yamtab.rx_fcsh;
                yamtab.rx_fcsh = chktabh[temp] ^ rxb;
            }
            set (rxinfo, yamtab.l2port);
        }
    }
}

/*** YAM TX *****************************************************************/

void yam_tx (void)
{
    word out;

    switch (yamtab.tx_state)
    {
    case DELAY:
        if (yamtab.tx_timer)
            return;

next_frm:
        yamtab.tx_fcsl  = 0x21;  /* start value of lower CRC byte (tx)     */
        yamtab.tx_fcsh  = 0xf3;  /* start value of upper CRC byte (tx)     */
        yamtab.tx_state = ACTIVE;

    case ACTIVE:
        if ((out = txget (yamtab.l2port)) < TXEOF)
        {
            yamtab.index   = yamtab.tx_fcsl;
            yamtab.tx_fcsl = chktabl[yamtab.index] ^ yamtab.tx_fcsh;
            yamtab.tx_fcsh = chktabh[yamtab.index] ^ (out & 0x0ff);
            break;
        }
        yamtab.tx_kicked = out != TXLAST;
        yamtab.tx_state  = TXCRC;
        yamtab.index     = yamtab.tx_fcsh ^ chktabl[yamtab.tx_fcsl];
        out  = chktabh[yamtab.tx_fcsl] ^ chktabl[yamtab.index]^0xff;
        break;

    case TXCRC:
        yamtab.tx_state  = FLUSH;
        out  = chktabh[yamtab.index] ^ 0xff;
        break;

    case FLUSH:
        yamtab.tx_timer  = TApar[yamtab.l2port] + clkstep;
        yamtab.tx_state  = TAIL;
        return;

    case TAIL:
        if (yamtab.tx_kicked)
            goto next_frm;

        if (yamtab.tx_timer)
            return;

        outp(yamtab.port_io + IER,0);        /* disable interrupt sources  */
        outp(yamtab.port_io + MCR, PTT_OFF); /* disable transmitter        */
        inp (yamtab.port_io + MSR);          /* clear pending ints         */
        inp (yamtab.port_io + RXB);
        outp(yamtab.port_io + IER,IER_RX);   /* enable receiver interrupts */
        yamtab.tx_state = IDLE;
        res (txinfo, yamtab.l2port);
        return;
    }

    /*  send char */
    outp(yamtab.port_io + TXB, (byte) out);
}

/*** YAM Interrupt **********************************************************/

void intr yam_int (void)
{
    byte msr;
    byte rxb;

    debug (YAM);

    msr = inp(yamtab.port_io + MSR);   /* clear pending modem (tx) int */
    rxb = inp(yamtab.port_io + RXB);   /* clear pending rx int         */

    if (yamtab.tx_state > DWAIT)
    {
        if(msr & TXRDY)
            yam_tx ();
    }
    else
        yam_rx (msr, rxb);

    eoi (yamtab.irq);
}

/*** YAM Init ***************************************************************/

void yam_init (void)
{
    yamtab.old_int = irq_init (yamtab.irq, yam_int);

    /*
     * communication to YAM takes place using a fixed baudrate of
     * 19k2. The speed on the air interface is determined by the
     *  initialisation of the FPGA on the modem (using YAMINIT)
     */
    outp(yamtab.port_io + IER, 0);              /* disable interrupt sources */
    outp(yamtab.port_io + LCR, LCR_DLAB|LCR_BIT8); /* set LCR to Baud,8,N,1  */
    outp(yamtab.port_io + DLSB, 6);             /* LSB of 115200/19200       */
    outp(yamtab.port_io + DMSB, 0);             /* MSB of 115200/19200       */
    outp(yamtab.port_io + LCR, LCR_BIT8);       /* Clear divisor access bit  */
    outp(yamtab.port_io + MCR, PTT_OFF);        /* Disable PTT               */

    inp (yamtab.port_io + RXB);                 /* purge the rx data buffer  */
    inp (yamtab.port_io + MSR);                 /* clear the modem status    */
    outp(yamtab.port_io + IER,IER_RX);          /* enable rx interrupts      */
}

/*** YAM Reset **************************************************************/

void yam_reset (void)
{
    outp (yamtab.port_io + IER, 0);
    outp (yamtab.port_io + MCR, PTT_OFF);

    irq_reset (yamtab.irq, yamtab.old_int);
}

/*** BPQ RX *****************************************************************/

void __cdecl __interrupt __far bpq_rx (
  unsigned _es, unsigned _ds,
  unsigned _di, unsigned _si,
  unsigned _bp, unsigned _sp,
  unsigned _bx, unsigned _dx,
  unsigned _cx, unsigned _ax,
  unsigned _ip, unsigned _cs,
  unsigned flags )
{
    /*
     * This is a call-back function of the packet driver,
     * for more information get a copy of the
     * PC/TCP Version 1.09 Packet Driver Specification by
     * FTP Software, Inc. I found my copy on the Internet,
     * it's free.
     */
    if (bpqtab.rx_size[bpqtab.rx_fill_nr] == 0)
    {
        if (_ax == 0)
        {
            /*
             * first call from the driver. Give a buffer address
             * in ES:DI
             */
            if(_cx > BPQ_BUFF_SIZE)
            {
                /* overflow, throw away */
                _es = 0;
                _di = 0;
            }
            else
            {
                /* buffer will fit, accept */
                _es = mac_ds;
                _di = bpqtab.rx_buf[bpqtab.rx_fill_nr];
            }
        }
        else
        {
                /* used in receive processing */
                if(_cx == 0) return; /* don't accept 0 length packets */
                idisable ();
                bpqtab.rx_size[bpqtab.rx_fill_nr] = _cx;
                /* next time fill the next buffer... */
                bpqtab.rx_fill_nr = (bpqtab.rx_fill_nr + 1) % BPQ_RX_BUF_NR;
                irestore ();
        }
    }
    else
    {
        if(_ax == 0)
        {
            /* busy, throw away */
            _es = 0;
            _di = 0;
        }
    }
}

void bpq_rx_process ()
{
    int  cnt, len, i;
    word in;
    int  nr;

    /*
     * called when rx_size[] not zero, bytes pending in buffer
     */

    /* check presence of miminum length */
    /* destination 6 bytes              */
    /* source      6 bytes              */
    /* frametype   2 bytes              */
    /* ax.25len+5  2 bytes              */
    /* payload     1 byte minimum       */
    nr = bpqtab.rx_read_nr;
    if(bpqtab.rx_size[nr] < (2*BPQ_ETHADDR_SIZE + 5))
    {
        /* discard packet */
        return;
    }

    if(bpqtab.rx_buf[nr][0] == 0x01)
    {
        /* possible multicast reception, check it, bail out on no match */
        if (bpqtab.rx_buf[nr][1] !=  'B') return;
        if (bpqtab.rx_buf[nr][2] !=  'P') return;
        if (bpqtab.rx_buf[nr][3] !=  'Q') return;
        if (bpqtab.rx_buf[nr][4] != 0x00) return;
        if (bpqtab.rx_buf[nr][5] != 0x00) return;
    }
    else
    {
        /* check if broadcast destination */
        for (cnt = 0; cnt < BPQ_ETHADDR_SIZE; cnt++)
        {
            /* broadcast destination, check it, bail out on no match */
            if (bpqtab.rx_buf[nr][cnt] != 0xff) return;
        }
    }

    /* skip over source address, we are not interested */
    cnt = (2 * BPQ_ETHADDR_SIZE);

    /* verify BPQ type */
    for(i = 0; i < BPQ_TYPE_SIZE; i++)
    {
        if(bpqtab.rx_buf[nr][cnt] != bpqtab.e_type[i])
        {
            /* not a BPQ type, discard packet */
            return;
        }
        cnt++;
    }

    /* retrieve length */
    len = bpqtab.rx_buf[nr][cnt++];
    len = len + (bpqtab.rx_buf[nr][cnt++] << 8);
    len -= 5; /* calculate AX.25 packet length */

    /* verify if the whole packet arrived */
    if(bpqtab.rx_size[nr] < (2*BPQ_ETHADDR_SIZE + 4 + len))
    {
        /* discard packet */
        return;
    }

    /* packet complete, cnt is index to first byte */
    while(len > 0)
    {
         in = bpqtab.rx_buf[nr][cnt++];
         rxput(bpqtab.l2port, in);
         len--;
    }
    rxeof(bpqtab.l2port, 0);
    l2call = true;
}

/*** BPQ TX *****************************************************************/

void bpq_tx_process (void)
{
    int  cnt1, cnt2, len, i;
    word out;

    if(bpqtab.tx_state == ACTIVE)
    {
        /* new transmission */
        if (bpqtab.driv_class == 1) /* only DIX Ethernet */
        {
            for ( cnt1 = 0, cnt2 = BPQ_ETHADDR_SIZE;
                  cnt1 < BPQ_ETHADDR_SIZE;
                  cnt1++, cnt2++)
            {
                /* destination */
                bpqtab.tx_buf[cnt1] = bpqtab.bpqaddr[cnt1];
                /* source */
                bpqtab.tx_buf[cnt2] = bpqtab.ethaddr[cnt1];
            }

            /* fill in frame type for BPQ */
            for(i = 0; i < BPQ_TYPE_SIZE; i++)
            {
                bpqtab.tx_buf[cnt2++] = bpqtab.e_type[i];
            }

            /* do length later, remember in cnt1 */
            cnt1 = cnt2;
            cnt2 += 2;

            len = 0;
            while((out = txget (bpqtab.l2port)) < TXEOF)
            {
                bpqtab.tx_buf[cnt2++] = (byte) out;
                len++;
            }

            if(len == 0)
            {
                /* do nothing and quit */
                bpqtab.tx_state = IDLE;
                return;
            }

            len += 5; /* calculate length, 5 more than AX frame */

            /* put length in little endian order */
            bpqtab.tx_buf[cnt1++] = (byte) len;
            bpqtab.tx_buf[cnt1++] = (byte) (len >> 8);

            bpqtab.tx_kicked = out != TXLAST;

            /* add dummy bytes if packet is too small */
            /* calculate packet length                */
            len = len - 5 + (2 * BPQ_ETHADDR_SIZE) + 4;

            /* National 8390 chips expects at least 60 bytes */
            if(len < 60) len = 60;
            bpqtab.tx_count = len;
        }
        else
        {
            /* do nothing and quit */
            bpqtab.tx_state = IDLE;
            return;
        }
    }

    /* send packet */
    inregs.h.ah = SEND_PACKET;
    inregs.x.bx = bpqtab.handle;
    segregs.ds  = mac_ds;
    inregs.x.si = bpqtab.tx_buf;
    inregs.x.cx = bpqtab.tx_count;
    _int86x(bpqtab.int_nr,&inregs,&outregs,&segregs);
    if( outregs.x.cflag == 0)
    {
        /* success */
        _disable();
        if(bpqtab.tx_kicked)
        {
                /* not TXLAST, on next run get new buffer */
                bpqtab.tx_state = ACTIVE;
        }
        else
        {
                /* TXLAST, ready until kicked again */
                bpqtab.tx_state = IDLE;
        }
        _enable();
    }
    else
    {
        /* On rext run do not refill */
        bpqtab.tx_state = DELAY;
    }
}

/*** BPQ RX - TX ************************************************************/

void bpq_rxtx(void)
{
    _disable ();

    if (!semaphore)
    {
        semaphore = true;
        _enable ();
        global_stack = set_stk (local_stack + STACK);

        if (bpqtab.rx_size[bpqtab.rx_read_nr])
        {
            bpq_rx_process ();
            bpqtab.rx_size[bpqtab.rx_read_nr] = 0; /* free to fill again */
            bpqtab.rx_read_nr = (bpqtab.rx_read_nr + 1) % BPQ_RX_BUF_NR;
        }
        if (bpqtab.tx_state)
        {
            bpq_tx_process ();
        }

        set_stk (global_stack);
        semaphore = false;
    }
    else
        _enable ();
}

/*** BPQ Init ***************************************************************/

void bpq_init (void)
{
    byte *p;
    int   i;
    int   rx_mode;  /* for loop variable */
    int   func = 1; /* basic functions if not changed */

    /* get data and code segment registers */
    _segread(&segregs);
    mac_cs = segregs.cs;
    mac_ds = segregs.ds;

    /* Set BPQ Ethernet type in e_type */
    bpqtab.e_type [0] = 0x08;
    bpqtab.e_type [1] = 0xff;

    /* get the driver type */
    inregs.h.ah = DRIVER_INFO;
    inregs.h.al = 0xff;
    inregs.x.bx = 0;
    _int86x(bpqtab.int_nr,&inregs,&outregs,&segregs);
    if( outregs.x.cflag == 0)
    {
        /* success */
        bpqtab.driv_class = outregs.h.ch;
        bpqtab.driv_type  = outregs.x.dx;
        bpqtab.driv_ifnum = outregs.h.cl;
        func = outregs.h.al;
    }
    else
    {
        /* driver_info failed, using defaults */
        bpqtab.driv_class = 1; /* DIX Ethernet */
        bpqtab.driv_type  = 0xFFFF;
        bpqtab.driv_ifnum = 0;
    }

    /* get access type */
    inregs.h.ah = ACCESS_TYPE;
    inregs.h.al = bpqtab.driv_class;
    inregs.x.bx = bpqtab.driv_type;
    inregs.h.dl = bpqtab.driv_ifnum;
    segregs.ds  = 0; /* mac_ds;        */
    inregs.x.si = 0; /* &bpqtab.e_type;  */
    inregs.x.cx = 0; /* BPQ_TYPE_SIZE;   */
    segregs.es  = mac_cs;
    inregs.x.di = (int) bpq_rx_callback;
    bpqtab.handle = _int86x(bpqtab.int_nr,&inregs,&outregs,&segregs);
    if( outregs.x.cflag == 0)
    {
        /* set flag to release this handle later */
        bpqtab.okay = true;
    }
    else
    {
        _enable ();
        cprintf("BPQ: driver_access failed\r\n");
        _disable();

        /* no need to release this handle later */
        bpqtab.okay = false;
        return;
    }

    /* get address */
    inregs.h.ah = GET_ADDRESS;
    inregs.x.bx = bpqtab.handle;
    segregs.es  = mac_ds;
    inregs.x.di = bpqtab.ethaddr;
    inregs.x.cx = BPQ_ETHADDR_SIZE;
    _int86x(bpqtab.int_nr,&inregs,&outregs,&segregs);
    /* no checking on errors */

    if((func & 0x02) != 0) /* check if extended functions are supported */
    {
        /* yes, extended functions supported */

        /* multicast receive from BPQ, try modes 5 or 6 */

        /* set receive mode */
        rx_mode = 5; /* try to set receive mode, either 5 or 6 */
        while(rx_mode < 7)
        {
            inregs.h.ah = SET_RCV_MODE;
            inregs.x.bx = bpqtab.handle;
            inregs.x.cx = rx_mode; /* try all multicast rx modes */
            _int86x(bpqtab.int_nr,&inregs,&outregs,&segregs);
            if( outregs.x.cflag == 0 )
            {
                /* success */
                break;
            }
            else
            {
                rx_mode++;
            }
        }

        if(rx_mode == 7)
        {
            _enable();
            cprintf("BPQ: Set receive mode failed (tried mode 5 and 6)\r\n");
            cprintf("BPQ: Fallback using broadcast packets...\r\n");
            _disable();
            for(i = 0; i < BPQ_ETHADDR_SIZE; i++)
                bpqtab.bpqaddr[i] = 0xff;
         }
    }
    else
    {
        _enable();
        cprintf("BPQ: Extended functions not supported by network driver\r\n");
        cprintf("BPQ: Fallback using broadcast packets...\r\n");
        _disable();
        for(i = 0; i < BPQ_ETHADDR_SIZE; i++)
            bpqtab.bpqaddr[i] = 0xff;
    }

    bpqtab.tx_state = IDLE;
    for(i = 0; i < BPQ_RX_BUF_NR; i++)
        bpqtab.rx_size[i] = 0;

    bpqtab.rx_fill_nr = 0; /* first buffer to fill is buffer 0 */
    bpqtab.rx_read_nr = 0; /* first buffer to read is buffer 0 */
}

/*** BPQ Reset **************************************************************/

void bpq_reset (void)
{
    /* release type */
    if(bpqtab.okay == true)
    {
        inregs.h.ah = RELEASE_TYPE;
        inregs.x.bx = bpqtab.handle;
        _int86x(bpqtab.int_nr,&inregs,&outregs,&segregs);
        if( outregs.x.cflag == 0)
        {
            /* success */
            bpqtab.okay = false;
        }
    }
}

/*** CRC SMACK **************************************************************/

void crc_smack (word *crc, byte data)
{
    static word crctab[] =

   {0x0000, 0xC0C1, 0xC181, 0x0140, 0xC301, 0x03C0, 0x0280, 0xC241,
    0xC601, 0x06C0, 0x0780, 0xC741, 0x0500, 0xC5C1, 0xC481, 0x0440,
    0xCC01, 0x0CC0, 0x0D80, 0xCD41, 0x0F00, 0xCFC1, 0xCE81, 0x0E40,
    0x0A00, 0xCAC1, 0xCB81, 0x0B40, 0xC901, 0x09C0, 0x0880, 0xC841,
    0xD801, 0x18C0, 0x1980, 0xD941, 0x1B00, 0xDBC1, 0xDA81, 0x1A40,
    0x1E00, 0xDEC1, 0xDF81, 0x1F40, 0xDD01, 0x1DC0, 0x1C80, 0xDC41,
    0x1400, 0xD4C1, 0xD581, 0x1540, 0xD701, 0x17C0, 0x1680, 0xD641,
    0xD201, 0x12C0, 0x1380, 0xD341, 0x1100, 0xD1C1, 0xD081, 0x1040,
    0xF001, 0x30C0, 0x3180, 0xF141, 0x3300, 0xF3C1, 0xF281, 0x3240,
    0x3600, 0xF6C1, 0xF781, 0x3740, 0xF501, 0x35C0, 0x3480, 0xF441,
    0x3C00, 0xFCC1, 0xFD81, 0x3D40, 0xFF01, 0x3FC0, 0x3E80, 0xFE41,
    0xFA01, 0x3AC0, 0x3B80, 0xFB41, 0x3900, 0xF9C1, 0xF881, 0x3840,
    0x2800, 0xE8C1, 0xE981, 0x2940, 0xEB01, 0x2BC0, 0x2A80, 0xEA41,
    0xEE01, 0x2EC0, 0x2F80, 0xEF41, 0x2D00, 0xEDC1, 0xEC81, 0x2C40,
    0xE401, 0x24C0, 0x2580, 0xE541, 0x2700, 0xE7C1, 0xE681, 0x2640,
    0x2200, 0xE2C1, 0xE381, 0x2340, 0xE101, 0x21C0, 0x2080, 0xE041,
    0xA001, 0x60C0, 0x6180, 0xA141, 0x6300, 0xA3C1, 0xA281, 0x6240,
    0x6600, 0xA6C1, 0xA781, 0x6740, 0xA501, 0x65C0, 0x6480, 0xA441,
    0x6C00, 0xACC1, 0xAD81, 0x6D40, 0xAF01, 0x6FC0, 0x6E80, 0xAE41,
    0xAA01, 0x6AC0, 0x6B80, 0xAB41, 0x6900, 0xA9C1, 0xA881, 0x6840,
    0x7800, 0xB8C1, 0xB981, 0x7940, 0xBB01, 0x7BC0, 0x7A80, 0xBA41,
    0xBE01, 0x7EC0, 0x7F80, 0xBF41, 0x7D00, 0xBDC1, 0xBC81, 0x7C40,
    0xB401, 0x74C0, 0x7580, 0xB541, 0x7700, 0xB7C1, 0xB681, 0x7640,
    0x7200, 0xB2C1, 0xB381, 0x7340, 0xB101, 0x71C0, 0x7080, 0xB041,
    0x5000, 0x90C1, 0x9181, 0x5140, 0x9301, 0x53C0, 0x5280, 0x9241,
    0x9601, 0x56C0, 0x5780, 0x9741, 0x5500, 0x95C1, 0x9481, 0x5440,
    0x9C01, 0x5CC0, 0x5D80, 0x9D41, 0x5F00, 0x9FC1, 0x9E81, 0x5E40,
    0x5A00, 0x9AC1, 0x9B81, 0x5B40, 0x9901, 0x59C0, 0x5880, 0x9841,
    0x8801, 0x48C0, 0x4980, 0x8941, 0x4B00, 0x8BC1, 0x8A81, 0x4A40,
    0x4E00, 0x8EC1, 0x8F81, 0x4F40, 0x8D01, 0x4DC0, 0x4C80, 0x8C41,
    0x4400, 0x84C1, 0x8581, 0x4540, 0x8701, 0x47C0, 0x4680, 0x8641,
    0x8201, 0x42C0, 0x4380, 0x8341, 0x4100, 0x81C1, 0x8081, 0x4040};

    *crc = HIBYTE (*crc) ^ crctab[LOBYTE (*crc) ^ data];
}

/*** CRC RMNC ***************************************************************/

void crc_rmnc (word *crc, byte data)
{
    static word crctab[] =

   {0x0F87, 0x1E0E, 0x2C95, 0x3D1C, 0x49A3, 0x582A, 0x6AB1, 0x7B38,
    0x83CF, 0x9246, 0xA0DD, 0xB154, 0xC5EB, 0xD462, 0xE6F9, 0xF770,
    0x1F06, 0x0E8F, 0x3C14, 0x2D9D, 0x5922, 0x48AB, 0x7A30, 0x6BB9,
    0x934E, 0x82C7, 0xB05C, 0xA1D5, 0xD56A, 0xC4E3, 0xF678, 0xE7F1,
    0x2E85, 0x3F0C, 0x0D97, 0x1C1E, 0x68A1, 0x7928, 0x4BB3, 0x5A3A,
    0xA2CD, 0xB344, 0x81DF, 0x9056, 0xE4E9, 0xF560, 0xC7FB, 0xD672,
    0x3E04, 0x2F8D, 0x1D16, 0x0C9F, 0x7820, 0x69A9, 0x5B32, 0x4ABB,
    0xB24C, 0xA3C5, 0x915E, 0x80D7, 0xF468, 0xE5E1, 0xD77A, 0xC6F3,
    0x4D83, 0x5C0A, 0x6E91, 0x7F18, 0x0BA7, 0x1A2E, 0x28B5, 0x393C,
    0xC1CB, 0xD042, 0xE2D9, 0xF350, 0x87EF, 0x9666, 0xA4FD, 0xB574,
    0x5D02, 0x4C8B, 0x7E10, 0x6F99, 0x1B26, 0x0AAF, 0x3834, 0x29BD,
    0xD14A, 0xC0C3, 0xF258, 0xE3D1, 0x976E, 0x86E7, 0xB47C, 0xA5F5,
    0x6C81, 0x7D08, 0x4F93, 0x5E1A, 0x2AA5, 0x3B2C, 0x09B7, 0x183E,
    0xE0C9, 0xF140, 0xC3DB, 0xD252, 0xA6ED, 0xB764, 0x85FF, 0x9476,
    0x7C00, 0x6D89, 0x5F12, 0x4E9B, 0x3A24, 0x2BAD, 0x1936, 0x08BF,
    0xF048, 0xE1C1, 0xD35A, 0xC2D3, 0xB66C, 0xA7E5, 0x957E, 0x84F7,
    0x8B8F, 0x9A06, 0xA89D, 0xB914, 0xCDAB, 0xDC22, 0xEEB9, 0xFF30,
    0x07C7, 0x164E, 0x24D5, 0x355C, 0x41E3, 0x506A, 0x62F1, 0x7378,
    0x9B0E, 0x8A87, 0xB81C, 0xA995, 0xDD2A, 0xCCA3, 0xFE38, 0xEFB1,
    0x1746, 0x06CF, 0x3454, 0x25DD, 0x5162, 0x40EB, 0x7270, 0x63F9,
    0xAA8D, 0xBB04, 0x899F, 0x9816, 0xECA9, 0xFD20, 0xCFBB, 0xDE32,
    0x26C5, 0x374C, 0x05D7, 0x145E, 0x60E1, 0x7168, 0x43F3, 0x527A,
    0xBA0C, 0xAB85, 0x991E, 0x8897, 0xFC28, 0xEDA1, 0xDF3A, 0xCEB3,
    0x3644, 0x27CD, 0x1556, 0x04DF, 0x7060, 0x61E9, 0x5372, 0x42FB,
    0xC98B, 0xD802, 0xEA99, 0xFB10, 0x8FAF, 0x9E26, 0xACBD, 0xBD34,
    0x45C3, 0x544A, 0x66D1, 0x7758, 0x03E7, 0x126E, 0x20F5, 0x317C,
    0xD90A, 0xC883, 0xFA18, 0xEB91, 0x9F2E, 0x8EA7, 0xBC3C, 0xADB5,
    0x5542, 0x44CB, 0x7650, 0x67D9, 0x1366, 0x02EF, 0x3074, 0x21FD,
    0xE889, 0xF900, 0xCB9B, 0xDA12, 0xAEAD, 0xBF24, 0x8DBF, 0x9C36,
    0x64C1, 0x7548, 0x47D3, 0x565A, 0x22E5, 0x336C, 0x01F7, 0x107E,
    0xF808, 0xE981, 0xDB1A, 0xCA93, 0xBE2C, 0xAFA5, 0x9D3E, 0x8CB7,
    0x7440, 0x65C9, 0x5752, 0x46DB, 0x3264, 0x23ED, 0x1176, 0x00FF};

    *crc = (*crc << 8) ^ crctab[HIBYTE (*crc) ^ data];
}

/*** KISS Interrupt *********************************************************/

void kiss_int (kiss_t *kiss)
{
    byte iir, rx_buf;
    buf_t *parbuf;
    word tx_buf;

    debug (KISS);

    eoi (kiss->irq);

    while (!((iir = RDCOM (2)) & 1))
    {
        switch (iir & 6)
        {
        case 2:
lost_int:   if (kiss->fesc)
            {
                WRCOM (0, kiss->fesc);
                kiss->fesc = 0;
            }
            else
            {
                switch (kiss->tx_state)
                {
                case 1:
                    switch (kiss->tx_mode)
                    {
                    case CRCOFF:
                        WRCOM (0, DATA);
                        break;

                    case CRCSMACK:
                        kiss->tx_crc = 0xA001;
                        WRCOM (0, SMACK);
                        break;

                    case CRCRMNC:
                        kiss->tx_crc = 0xDEFD;
                        WRCOM (0, RMNC);
                    }
                    kiss->tx_state = 2;
                    break;

                case 2:
                    if ((tx_buf = txget (kiss->l2port)) < TXEOF)
                    {
                        switch (kiss->tx_mode)
                        {
                        case CRCSMACK:
                            crc_smack (&kiss->tx_crc, (byte) tx_buf);
                            break;

                        case CRCRMNC:
                            crc_rmnc (&kiss->tx_crc, (byte) tx_buf);
                        }

slip_out:               if (tx_buf == FEND || tx_buf == FESC)
                        {
                            kiss->fesc = tx_buf == FEND ? TFEND : TFESC;
                            tx_buf = FESC;
                        }

                        WRCOM (0, tx_buf);
                    }
                    else
                    {
                        kiss->tx_kicked = tx_buf != TXLAST;

                        if (kiss->tx_mode)
                        {
                            tx_buf =   kiss->tx_mode == CRCSMACK
                                     ? LOBYTE (kiss->tx_crc)
                                     : HIBYTE (kiss->tx_crc);
                            kiss->tx_state = 3;
                            goto slip_out;
                        }
                        else
                        {
                            WRCOM (0, FEND);
                            kiss->tx_state = 5;
                        }
                    }
                    break;

                case 3:
                    tx_buf =   kiss->tx_mode == CRCSMACK
                             ? HIBYTE (kiss->tx_crc)
                             : LOBYTE (kiss->tx_crc);
                    kiss->tx_state = 4;
                    goto slip_out;

                case 4:
                    WRCOM (0, FEND);
                    kiss->tx_state = 5;

                    if (!kiss->crc_ok)
                        switch (kiss->tx_mode)
                        {
                        case CRCSMACK:
                            kiss->tx_mode = CRCRMNC;
                            break;

                        case CRCRMNC:
                            kiss->tx_mode = CRCOFF;
                        }
                    break;

                case 5:
                    if (!kiss->parlst)
                    {
kiss_cont:              if (kiss->tx_kicked)
                        {
                            WRCOM (0, FEND);
                            kiss->tx_state = 1;
                        }
                        else
                        {
                            res (txinfo, kiss->l2port);
                            kiss->tx_state = 0;
                        }
                        break;
                    }
                    kiss->tx_state = 6;

                case 6:
                    if (kiss->bufp == 4)
                    {
                        kiss->bufp = 0;

                        parbuf = kiss->parlst;
                        kiss->parlst = parbuf->next;
                        dealoc (parbuf);

                        if (!kiss->parlst)
                            goto kiss_cont;
                    }

                    WRCOM (0, kiss->parlst->data[kiss->bufp++]);
                }
            }
            break;

        case 4:
            rx_buf = RDCOM (0);

            set (rxinfo, kiss->l2port);
            kiss->dcd_timer = not0 (Cpar[kiss->l2port]);

            if (!(RDCOM (5) & 0x1E))
            {
                switch (kiss->rx_state)
                {
                case 0:
                    if (rx_buf == FEND)
                        kiss->rx_state = 1;
                    break;

                case 1:
                    switch (rx_buf)
                    {
                    case FEND:
                        break;

                    case DATA:
                    case SMACK:
                    case RMNC:
                        if (   !kiss->crc_ok
                            ||    kiss->tx_mode == CRCSMACK
                               && rx_buf        == SMACK
                            ||    kiss->tx_mode == CRCRMNC
                               && rx_buf        == RMNC)
                        {
                            switch (rx_buf)
                            {
                            case DATA:
                                kiss->rx_mode = CRCOFF;
                                break;

                            case SMACK:
                                kiss->rx_mode = CRCSMACK;
                                kiss->rx_crc  = 0xA001;
                                break;

                            case RMNC:
                                kiss->rx_mode = CRCRMNC;
                                kiss->rx_crc  = 0xDEFD;
                            }

                            kiss->rx_len   = 0;
                            kiss->rx_state = 2;
                            break;
                        }

                    default:
                        kiss->rx_state = 0;
                    }
                    break;

                case 2:
                    switch (rx_buf)
                    {
                    case FEND:
                        if (kiss->rx_mode)
                        {
                            if (   kiss->rx_len < MIN_LEN+2
                                ||    kiss->rx_mode == CRCSMACK
                                   && kiss->rx_crc  != 0
                                ||    kiss->rx_mode == CRCRMNC
                                   && kiss->rx_crc  != 0x7070)
                                goto kiss_err;

                            kiss->crc_ok  = true;
                            kiss->tx_mode = kiss->rx_mode;

                            rxeof (kiss->l2port, 2);
                        }
                        else
                        {
                            if (kiss->rx_len < MIN_LEN)
                                goto kiss_err;

                            rxeof (kiss->l2port, 0);
                        }

                        l2call = true;
                        kiss->rx_state = kiss->crc_ok ? 1 : 0;
                        break;

                    case FESC:
                        kiss->rx_state = 3;
                        break;

                    default:
kiss_put:               kiss->rx_len++;

                        switch (kiss->rx_mode)
                        {
                        case CRCSMACK:
                            crc_smack (&kiss->rx_crc, rx_buf);
                            break;

                        case CRCRMNC:
                            crc_rmnc (&kiss->rx_crc, rx_buf);
                        }

                        if (rxput (kiss->l2port, rx_buf))
                            goto kiss_err;
                    }
                    break;

                case 3:
                    if (rx_buf == TFEND || rx_buf == TFESC)
                    {
                        rx_buf = rx_buf == TFEND ? FEND : FESC;
                        kiss->rx_state = 2;
                        goto kiss_put;
                    }
                    else
                        goto kiss_err;
                }
            }
            else
            {
kiss_err:       rxdiscard (kiss->l2port);
                kiss->rx_state = 0;
                kiss->errors++;
            }
        }
    }

    if (kiss->tx_state && (RDCOM (5) & 0x20))
        goto lost_int;
}

void intr kiss_int0 (void) { kiss_int (&kisstab[0]); }

void intr kiss_int1 (void) { kiss_int (&kisstab[1]); }

void intr kiss_int2 (void) { kiss_int (&kisstab[2]); }

void intr kiss_int3 (void) { kiss_int (&kisstab[3]); }

/*** KISS Init **************************************************************/

procfp hook[] = {kiss_int0, kiss_int1, kiss_int2, kiss_int3};

void kiss_init (void)
{
    kiss_t *kiss;
    word tconst;

    for (kiss = kisstab; kiss < kissend; kiss++)
    {
        kiss->tx_mode = CRCSMACK;

        WRCOM (1, 0);
        WRCOM (4, 0);

        kiss->old_int = irq_init (kiss->irq, hook[kiss - kisstab]);

        tconst = 115200L / kiss->baud;

        WRCOM (3, 0x83);
        WRCOM (0, LOBYTE (tconst));
        WRCOM (1, HIBYTE (tconst));
        WRCOM (3, 0x03);

        while (RDCOM (5) & 1)
            RDCOM (0);

        WRCOM (4, 0x0B);
        WRCOM (1, 0x03);

        l1par (kiss->l2port, TPAR,  defTp);
        l1par (kiss->l2port, PPAR,  defPp);
        l1par (kiss->l2port, WPAR,  defWp);
        l1par (kiss->l2port, DPAR,  defDp);
    }
}

/*** KISS Reset *************************************************************/

void kiss_reset (void)
{
    kiss_t *kiss;

    for (kiss = kisstab; kiss < kissend; kiss++)
    {
        WRCOM (1, 0);
        WRCOM (4, 0);

        irq_reset (kiss->irq, kiss->old_int);
    }
}

/*** Interrupt Handlers *****************************************************/

void intr int8 (void)
{
    static word tick_cnt = 1,
                int8_cnt = 1;
    modem_t *modem;

    debug (MODEM);

    for (modem = modemtab; modem < modemend; modem++)
    {
        if (--modem->baud_cnt == 0)
        {
            modem->baud_cnt = modem->baud_delay;

            if (--modem->send_cnt == 0)
            {
                modem->send_cnt = 3;
                modem_tx (modem);               /* 1200/s */
            }

            if (modem->tx_state <= DWAIT || Dpar[modem->l2port])
                modem_rx (modem);               /* 3600/s */
        }
    }

    _enable ();
    eoi (0);

    if (clockdev == MODEM && --tick_cnt == 0)
    {
        tick_cnt = tick_delay;
        ticks += 2;                             /* 50/s */

        timer (2);
    }

    if (--int8_cnt == 0)
    {
        int8_cnt = int8_delay;
        chain (old_int8);                       /* 18.2/s */
    }

    /*
     * Count the number of int8 interrupts. Used in the RTC ISR to
     * keep the PC timer at the correct speed
     */
    clock_int8_count++;
}

void intr int1C (void)
{
    static byte cnt;

    debug (LOOP);

    if (clockdev == LOOP)
    {
        ticks += cnt = cnt == 5 ? 6 : 5;

        _enable ();
        timer (cnt);
    }

    chain (old_int1C);
}

/*** Init Interrupts ********************************************************/

void init_intr (void)
{
    _disable ();

    if(modems)
      old_int8   = _dos_getvect (0x08);
    old_int1C  = _dos_getvect (0x1C);
    old_defInt = _dos_getvect (defInt);

    if(modems)
      _dos_setvect (0x08, int8);
    _dos_setvect (0x1C, int1C);
    _dos_setvect (defInt, iface_mac);

    if (modems)    modem_init ();
    if (parports)  par_init ();
    if (yamports)  yam_init ();
    if (bpqports)  bpq_init ();
    if (sccs)      scc_init ();
    if (kissports) kiss_init ();

    if (isat ())
        rtc_init ();

    if (l1ports && !modems)
        irq_prio (2);
    _enable ();
}

/*** Reset Interrupts *******************************************************/

void _cdecl reset_intr (void)
{
    _disable ();

    if (l1ports && !modems)
        irq_prio (0);

    if (isat ())
        rtc_reset ();

    if (kissports) kiss_reset ();
    if (sccs)      scc_reset ();
    if (parports)  par_reset ();
    if (yamports)  yam_reset ();
    if (bpqports)  bpq_reset ();
    if (modems)    modem_reset ();

    if(modems)
      _dos_setvect (0x08,   old_int8);
    _dos_setvect (0x1C,   old_int1C);
    _dos_setvect (defInt, old_defInt);

    _enable ();
}

/*** Memory Sizes ***********************************************************/

byte *minmem (void)
{
    return bufbgn;
}

byte *maxmem (void)
{
    return bufend - 1;
}

/*** Keep it resident *******************************************************/

void mac_keep (word retcode)
{
    bytefp endp = (bytefp) bufend;

    _dos_freemem (*(wordfp) (((dword) _psp << 16) + 0x2C));
    _dos_keep (retcode,
              (word) ((dword) FP_OFF (endp)+15 >> 4) + FP_SEG (endp) - _psp);
}

/*** AX25_MAC loaded? *******************************************************/

byte hdr_mac[] = "??AX25_MAC";

bool loaded (void)
{
    byte vector, *p1;
    bytefp p2, handler;

    for (vector = 0xFF; vector >= 0x40; vector--)
    {
        handler = (bytefp) _dos_getvect (vector);

        p1 = hdr_mac;

        for (p2 = handler; *p1 == '?' || *p1 == *p2; p2++)
            if (!*++p1)
            {
                defInt = vector;
                return true;
            }
    }

    return false;
}

/*** Error Message **********************************************************/

enum {ERR_OPT,    ERR_PORT,   ERR_MANY, ERR_DETECT, ERR_ADDR, ERR_INT,
      ERR_NOLOAD, ERR_LOADED, ERR_VER,  ERR_BAUD,   ERR_FILE, ERR_NUM,
      ERR_IRQ,    ERR_MODE,   ERR_BUF,  ERR_CHAN,   ERR_MTASK, ERR_YAMINIT};

char *msg[] = {"Invalid option (-H for help)",
               "Invalid port",
               "Too many ports",
               "Port not present",
               "Port address out of range",
               "Interrupt out of range",
               "AX25_MAC not loaded",
               "AX25_MAC already loaded",
               "Invalid version number",
               "Invalid baud rate",
               "Can't open file",
               "Invalid numerical format",
               "IRQ out of range",
               "Invalid clock mode",
               "Number of buffers out of range",
               "Number of channels out of range",
               "Serial modems don't work in this system",
               "YAM not configured. Run YAMINIT to initialize"};

void errmsg (byte err)
{
    cprintf ("\nError: %s\r\n", msg[err]);
    exit (EXIT_FAILURE);
}

/*** Signon *****************************************************************/

void signon (void)
{
    byte port;

    cprintf ("Ŀ\r\n"
             "   AX25_MAC v"__VER__" ("__DATE__")      \r\n"
             "              by PE1DNN               \r\n"
             "Ĵ\r\n"
             " Contains portions of TFPCX by DG0FT  \r\n"
             " Published under GNU GPL License.     \r\n");
    cprintf ("Ĵ\r\n"
             " %u Port(s), AX25_MAC-Int %X           \r\n"
             " %-3u Buffers(s)                       \r\n",
             l1ports, defInt, bufcnt);

    if (l1ports)
    {
        cputs ("Ĵ\r\n");

        for (port = 0; port < l1ports; port++)
            cprintf (" %u: %s (%X/%02u), %5u Bd, %-7s  \r\n",
                     port, l1info (port, L1NAME), l1info (port, L1ADDR),
                           l1info (port, L1IRQ),  l1info (port, L1BAUD),
                           l1info (port, L1TYPE));
    }

    cputs ("\r\n");
}

/*** Token ******************************************************************/

char *token (char *str)
{
    static char *pos;

    if (str)
        pos = str;
    else
        str = pos;

    while (*pos && *pos != ':')
        pos++;

    if (*pos)
        *pos++ = '\0';

    return str;
}

/*** Number *****************************************************************/

word number (char *str, byte base)
{
    dword value = 0;
    byte digit;

    if (*(str = token (str)))
    {
        while (*str)
        {
            if ((digit = *str++ - '0') > 9)
                error (ERR_NUM, (digit -= 7) < 10 || digit >= base);

            error (ERR_NUM, (value = value * base + digit) > 0xFFFF);
        }

        error (ERR_NUM, !value);
    }

    return (word) value;
}

/*** Get Address ************************************************************/

word getaddr (word defaddr)
{
    word addr;

    if (addr = number (NULL, 16))
    {
        error (ERR_ADDR,    addr < 0x100
                         || addr > 0x3F8
                         || addr & 7);
    }
    else
        error (ERR_DETECT, !(addr = defaddr));

    return addr;
}

/*** Get IRQ ****************************************************************/

byte getirq (byte defirq)
{
    byte irq;

    if (irq = number (NULL, 10))
    {
        error (ERR_IRQ,    irq  < 2
                        || irq == 6
                        || irq == 8
                        || irq == 13
                        || irq  > 15);
        if (isat ())
        {
            if (irq == 2) irq = 9;
        }
        else
            error (ERR_IRQ, irq > 7);
    }
    else
        irq = defirq;

    return irq;
}

/*** Add Port ***************************************************************/

byte addport (byte major, void *minor)
{
    error (ERR_MANY, l1ports == L1PORTS);

    l1port[l1ports].major = lastdev = major;
    l1port[l1ports].minor = minor;

    return l1ports++;
}

/*** Set Timer **************************************************************/

void settimer (byte major, byte prio, byte step)
{
    if (prio > clkprio)
    {
        clockdev = major;
        clkprio  = prio;
        clkstep  = step;
    }
}

/*** Modem Config ***********************************************************/

void modem_cfg (char *arg)
{
    modem_t *modem;

    error (ERR_MANY, modems++ == MODEMS);

    modem->l2port = addport (MODEM, modem = modemend++);

    error (ERR_PORT,    (modem->rs232 = strncmp (arg, "LPT", 3))
                     &&                 strncmp (arg, "COM", 3)
                     || (modem->port = number (arg+3, 10) - 1) > 3
                     || arg[4]);

    modem->port_out = getaddr (BIOSBASE[modem->port]);

    if (modem->rs232)
    {
        modem->port_inp = 2 + (modem->port_out += 4);
        modem->ptt_on   = 2;
        modem->ptt_off  = 1;
        modem->data_out = 1;
        modem->data_inp = 0x10;
    }
    else
    {
        modem->port_inp = 1 + modem->port_out;
        modem->ptt_on   = 0xBF;
        modem->ptt_off  = 0x3F;
        modem->data_out = 0x40;
        modem->data_inp = 0x80;
    }

    settimer (MODEM, 100, 2);
}

/*** PAR Config *************************************************************/

void par_cfg (char *arg)
{
    word counter;

    error (ERR_MANY, parports++);

    partab.l2port = addport (PAR, &partab);

    error (ERR_PORT,    strncmp (arg, "PAR", 3)
                     || (partab.port = number (arg+3, 10) - 1) > 3
                     || arg[4]);

    partab.port_out = getaddr (LPTBASE[partab.port]);
    partab.port_inp = partab.port_out + 1;

    partab.irq = getirq (7);

    outp (partab.port_out, PTT);
}

/*** YAM Config *************************************************************/

void yam_cfg (char *arg)
{
    error (ERR_MANY, yamports++);

    yamtab.l2port = addport (YAM, &yamtab);

    error (ERR_PORT,    strncmp (arg, "YAM", 3)
                     || (yamtab.port = number (arg+3, 10) - 1) > 3
                     || arg[4]);

    yamtab.port_io = getaddr (COMBASE[yamtab.port]);

    yamtab.irq = getirq ((byte) (4 - (yamtab.port & 1)));

    error (ERR_YAMINIT, !(inp(yamtab.port_io + MSR) & FPGA_CFG));

    outp (yamtab.port_io + MCR, PTT_OFF);
}

/*** BPQ Config *************************************************************/

void bpq_cfg (char *arg)
{
    char         int_nr;
    char far    *fstr;
    static char *driver_id = "PKT DRVR";

    error (ERR_MANY, bpqports++);

    bpqtab.l2port = addport (BPQ, &bpqtab);

    error (ERR_PORT,    strncmp (arg, "BPQ", 3));
    if(arg[3] != '\0')
    {
        error (ERR_IRQ, ((bpqtab.int_nr = number (arg+3, 16)) < 0x60)
                        || arg[5]);
    }
    else
    {
        /* search for packet driver */
        bpqtab.int_nr = 0;
        for(int_nr = 0x60; int_nr <= 0x80; int_nr++)
        {
            fstr = (char far *) _dos_getvect (int_nr);
            if(_fstrcmp((char far *)driver_id, &fstr[3]) == 0)
            {
                bpqtab.int_nr = int_nr;
                break;
            }
        }
        error (ERR_IRQ, bpqtab.int_nr == 0);
    }

    error (ERR_IRQ, bpqtab.int_nr > 0x80 || bpqtab.int_nr < 0x60);

    /* check if there is a packet driver */
    fstr = (char far *) _dos_getvect (bpqtab.int_nr);
    error (ERR_DETECT, _fstrcmp((char far *)driver_id, &fstr[3]));

    /* Packet driver okay! */

    /* program multicast address */
    bpqtab.bpqaddr[0] = 0x01;
    bpqtab.bpqaddr[1] = 'B';
    bpqtab.bpqaddr[2] = 'P';
    bpqtab.bpqaddr[3] = 'Q';
    bpqtab.bpqaddr[4] = 0x00;
    bpqtab.bpqaddr[5] = 0x00;
}

/*** SCC Config *************************************************************/

sccboard_t sccboard[] = {{0x300, 7, "2222", {1, 0, 3, 2}, 16},  /* DSCC */
                         {0x150, 3, "4445", {3, 1, 7, 5}, -1},  /* OSCC */
                         {0x300, 7, "1103", {0, 1, 2, 3}, 4}};  /* USCC */

void scc_cfg (char *arg)
{
    byte chan, lastchan, modem;
    sccboard_t *board;
    word sccbase;
    char *modems;
    scc_t *scc;

    error (ERR_MANY, sccs);
    error (ERR_PORT, strcmp (token (arg+1), "SCC"));

    board = &sccboard[scctype];

    sccbase = getaddr (board->base);

    sccirq = getirq (board->irq);

    if (!*(modems = token (NULL)))
        modems = board->modems;

    for (chan = 0, scc = scctab; chan < SCCCHANS; chan++, scc++)
    {
        scc->chan = chan;
        scc->data = sccbase + board->data[chan];
        scc->ctrl = scc->data + board->ctrl;

        if (*modems)
        {
            error (ERR_MODE, (modem = *modems++ - '0') > CLKCHAN);

            if (modem)
            {
                if (modem == CLKCHAN)
                {
                    error (ERR_MODE, scctype != OSCC);
                    settimer (SCC, 75, 2);
                    sccclkchan = scc;
                }
                else
                {
                    scc->l2port = addport (SCC, scc);
                    scc->modem = modem;
                    sccs++;
                }

                lastchan = chan;
            }
        }
    }

    error (ERR_MODE, *modems || !sccs);

    sccend = &scctab[(lastchan + 2) & 0xFE];

    switch (scctype)
    {
    case USCC: sccclkchan = &scctab[1];
    case DSCC: settimer (SCC, 75, 2);
    }
}

/*** KISS Config ************************************************************/

void kiss_cfg (char *arg)
{
    kiss_t *kiss;

    error (ERR_MANY, kissports++ == KISSPORTS);

    kiss->l2port = addport (KISS, kiss = kissend++);

    error (ERR_PORT,    strncmp (arg, "KISS", 4)
                     || (kiss->port = number (arg+4, 10) - 1) > 3
                     || arg[5]);

    kiss->base = getaddr (COMBASE[kiss->port]);

    kiss->irq = getirq ((byte) (4 - (kiss->port & 1)));
}

/*** Baud Config ************************************************************/

void baud_cfg (char *arg)
{
    word baud, maxbaud = 0;
    modem_t *modem;
    port_t *port;
    scc_t *scc;
    byte prt;

    for (prt = 0, port = l1port; prt < l1ports; prt++, port++, arg = NULL)
    {
        baud = number (arg, 10);

        switch (port->major)
        {
        case MODEM:
            if (!baud)
                baud = 1200;

            error (ERR_BAUD,    baud !=  300 && baud != 1200
                             && baud != 2400 && baud != 4800);

            if ((port->modem->baud = baud) > maxbaud)
                maxbaud = baud;

            break;

        case PAR:
            if (!baud)
                baud = 9600;

            error (ERR_BAUD, baud != 9600 && baud != 19200);

            partab.baud = baud;

            break;

        case YAM:
            if (!baud)
                baud = 9600;

            error (ERR_BAUD, (baud < 45 || baud > 19200));

            yamtab.baud = baud;

            break;

        case SCC:
            scc = port->scc;

            if (!baud)
                baud =    scc->modem == SOFTCLK
                       || scc->modem == PA0HZP ? 1200 : 9600;

            error (ERR_BAUD,    scc->modem != DF9IC
                             && (baud < 50 || baud > 38400));

            scc->baud = baud;

            break;

        case KISS:
            if (!baud)
                baud = 9600;

            error (ERR_BAUD,    baud !=  2400 && baud !=  4800
                             && baud !=  9600 && baud != 19200
                             && baud != 38400 && baud != 57600);

            port->kiss->baud = baud;
        }
    }

    for (modem = modemtab; modem < modemend; modem++)
    {
        modem->baud_delay =
        modem->baud_cnt   = maxbaud / modem->baud;
        modem->send_cnt   = 3;
    }

    switch (maxbaud)
    {
    case  300: pit0_const = 1325; int8_delay =  49; tick_delay =  18; break;
    case 1200: pit0_const =  331; int8_delay = 198; tick_delay =  72; break;
    case 2400: pit0_const =  166; int8_delay = 395; tick_delay = 144; break;
    case 4800: pit0_const =   83; int8_delay = 790; tick_delay = 288; break;
    }
}

/*** Help *******************************************************************/

char help[] =

"\n"
"Usage: AX25_MAC [ -N ] [ <load options> | -U ]\r\n"
"\n"
"<general options>              <legend>\r\n"
"  -N no messages                 [] optional\r\n"
"  -U unload                      |  alternative\r\n"
"                                 x  hex digit\r\n"
"                                 n  dec digit\r\n"
"<load options>\r\n"
"  -P<port>[:xxx:nn:nnnn]  packet port [addr:IRQ:<clock>]\r\n"
"  -Bnnnn[:nnnn ...]       baud rate (1 number/port)\r\n"
"\r\n"
"  -F[file]  read init file\r\n"
"  -D        debug mode\r\n"
"  -C[xx]    show DCD [color]\r\n"
"  -Ixx      AX25_MAC interrupt\r\n"
"  -L        interLock - one TX at a time (only for half-duplex ports)\r\n"
"  -BU[nnnn] number of buffers\r\n"
"<port>  COMn | LPTn | PARn | YAMn | BPQnn | KISSn | DSCC | OSCC | USCC\r\n"
"        (n = 1-4, for BPQ n=  60-80)\r\n"
"<clock> 0 = disable    2 = hardclock    4 = PA0HZP port  (1 digit/\r\n"
"        1 = softclock  3 = DF9IC modem  5 = PA0HZP timer  channel)\r\n";

/*** Parameter processing ***************************************************/

void mac_out(char *string)
{
    char *command;
    char *port;
    char *par;
    int  port_nr;
    int  par_nr;
    int  par_max;

    command = strtok(string," \t");
    if(command == NULL)
        return;

    switch(command[0]) {
    case '@':
            switch(command[1]) {
            case 'C':
                    par_max = 63;
                    break;
            case 'D':
                    par_max = 1;
                    break;
            case 'T':
                    if(command[2] == 'A')
                    {
                        par_max = 32767;
                        break;
                    }
                    /* else fall-thru */
            default:
                    cprintf("Warning: "
                            "ignoring unknown extended command '%s'\r\n",
                            command);
                    return;
            }
            break;
    case 'P':
    case 'T':
    case 'W':
            par_max = 255;
            break;
    case 'X':
            par_max = 1;
            break;
    default:
            cprintf("Warning: ignoring unknown command '%s'\r\n", command);
            return;
    }

    port = strtok(NULL," \t:");
    if(port == NULL)
    {
        cprintf("Warning: ignoring command '%s'; "
                "missing port number\r\n", command);
        return;
    }

    port_nr = atoi(port);
    if((port_nr < 0) || (port_nr > 7))
    {
        cprintf("Warning: ignoring command '%s'; "
              "port number out of range (0-7, is '%d')\r\n", command, port_nr);
        return;
    }

    par = strtok(NULL," \t");
    if(par == NULL)
    {
        cprintf("Warning: ignoring command '%s' for port '%s'; "
                "missing parameter number\r\n", command, port);
        return;
    }

    par_nr = atoi(par);

    if((par_nr < 0) || (par_nr > par_max))
    {
        cprintf("Warning: ignoring command '%s' for port '%s'; "
                "parameter out of range (0-%d, is %d)\r\n",
                command, port, par_max, par_nr);
                return;
    }

    switch(command[0]) {
    case '@':
            switch(command[1]) {
            case 'C':
                    Cpar[port_nr] = par_nr;
                    break;
            case 'D':
                    Dpar[port_nr] = par_nr;
                    break;
            case 'T':
                    TApar[port_nr] = par_nr;
                    break;
            default:
                    return;
            }
            break;
    case 'P':
            Ppar[port_nr] = par_nr;
            break;
    case 'T':
            Tpar[port_nr] = par_nr;
            break;
    case 'W':
            Wpar[port_nr] = par_nr;
            break;
    case 'X':
            Xpar[port_nr] = par_nr;
            break;
    default:
            return;
    }
}

/*** Read Init **************************************************************/

void read_init (word handle)
{
    bool empty = true, ignore = false;
    word spaces = 0;
    byte ch;
    char string[80];
    int  pos = 0;

    if (logo)
        cputs ("\nReading ...\r\n\n");

    while (read (handle, &ch, 1) == 1)
    {
        if (ch == '\n')
        {
            if (!empty)
            {
                string[pos] = '\0';
                pos = 0;
                mac_out (string);
            }

            empty  = true;
            ignore = false;
            spaces = 0;
        }
        else
            if (!ignore)
                switch (ch)
                {
                case ' ':
                case '\t':
                    spaces++;
                    break;

                case '#':
                case ';':
                    ignore = true;
                    break;

                case '^':
                    if (empty)
                        break;

                default:
                    if (empty)
                    {
                        empty  = false;
                        spaces = 0;
                    }
                    else
                        for (; spaces; spaces--)
                            if(pos < 79)
                                string[pos++] = ' ';

                    if(pos < 79)
                        string[pos++] = ch;
                }
    }

    close (handle);
}

/*** Set Envp ***************************************************************/

void _cdecl _setenvp (void)         /* suppress copying environment */
{
}

/*** Main *******************************************************************/

void main (word argc, char **argv)
{
    char ch, *baudarg = "", *initfile = NULL;
    word bufsize, handle;

    while (*++argv)
    {
        error (ERR_OPT, *strupr((*argv)++) != '-');

        switch (*(*argv)++)
        {
        case 'B':
            if (**argv == 'U')
                bufcnt = number (*argv+1, 10);
            else
                baudarg = *argv;
            break;

        case 'C':
            attrib  = (byte) number (*argv, 16);
            carrier = true;
            break;

        case 'D':
            debugdev = lastdev;
            break;

        case 'F':
            initfile = **argv ? *argv : "ax25_mac.ini";
            break;

        case 'H':
        case '?':
            cputs (help);
            exit (EXIT_SUCCESS);

        case 'I':
            defInt = number (*argv, 16);
            error (ERR_INT, defInt < 0x40 || 0xFF < defInt);
            break;

        case 'L':
            interlock = true;
            break;

        case 'N':
            logo = false;
            break;

        case 'P':
            switch (**argv)
            {
            case 'U': scctype++;
            case 'O': scctype++;
            case 'D': scc_cfg   (*argv); break;
            case 'K': kiss_cfg  (*argv); break;
            case 'P': par_cfg   (*argv); break;
            case 'Y': yam_cfg   (*argv); break;
            case 'B': bpq_cfg   (*argv); break;
            default:  modem_cfg (*argv);
            }
            break;

        case 'U':
            error (ERR_NOLOAD, !loaded ());
            error (ERR_VER, call_mac (0xFE00) != VERSION);
            _dos_freemem (call_mac (0xFF00));

            if (logo)
                cputs ("\nAX25_MAC unloaded\r\n");
            exit (EXIT_SUCCESS);

        default:
            errmsg (ERR_OPT);
        }
    }

    if (isat () && !mtasker ())
        settimer (RTC, 64, 2);

    error (ERR_LOADED, loaded ());

    if (!bufcnt)
        bufcnt = _memmax () / sizeof(pool_t);

    error (ERR_BUF,    bufcnt < 40
                    || bufcnt > 0xFFFF / sizeof(pool_t)
                    || !(bufbgn = malloc (bufsize = bufcnt * sizeof(pool_t))));
    bufend = bufbgn + bufsize;

    baud_cfg (baudarg);

    error (ERR_MTASK, modems && mtasker ());

    error (ERR_DETECT, sccs && !scc_detect ());

    if (initfile)
        error (ERR_FILE, (handle = open (initfile, O_RDONLY | O_TEXT)) == -1);

    if (logo)
        signon ();

    mainf ();
    init_intr ();

    if (initfile)
        read_init (handle);

    mac_keep (EXIT_SUCCESS);
}

/****************************************************************************/
