/*
 *  Copyright (C) 2000-2002  Henk de Groot - PE1DNN
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 */
/*
 * This file contains lists of calls, used for all kinds of purposes,
 * like a list of calls that are blocked and a list of digi-owners.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _LINUX_
#include "linux.h"
#endif
#include "digicons.h"
#include "digitype.h"
#include "mac_if.h"
#include "genlib.h"
#include "read_ini.h"
#include "call.h"
#include "output.h"

/*
 * Here are the lists is used to store call signs.
 */
static call_t *digi_owner_p      = NULL; /* digi maintainers              */
static call_t *digi_block_p      = NULL; /* blocked calls for digipeating */
static call_t *digi_via_block_p  = NULL; /* blocked calls for digipeating */
static call_t *digi_allow_to_p[MAX_PORTS] = {
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
                                         /* allowed "to" calls for digi   */
static call_t *digi_allow_from_p[MAX_PORTS] = {
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
                                         /* allowed "from" calls for digi */
static call_t *digi_msg_block_p  = NULL; /* blocked calls for query       */

/*
 * Add call(s) from a command from the .ini file to a list.
 *
 * - The first argument is a string that contains a list of calls.
 * - The second argument is the the text for the type of command - used for
 *   error messages.
 * - The third argument is a pointer to the list pointer which points to the
 *   list on which the calls shall be added.
 *
 * Warning, the function uses strtok on calls_p!! (strtok modifies the
 * string)
 */
static void common_add_call(char *calls_p, char *type, call_t **call_list_pp)
{
    call_t *call_p;
    char   *call;

    if(calls_p == NULL)
    {
        say("No calls found while handling %s: line at line %d\r\n",
                        type, linecount);
        return;
    }

    call = strtok(calls_p,", \t\r\n");

    while(call != NULL)
    {
        if(strlen(call) < CALL_LENGTH)
        {
            call_p = (call_t*) malloc(sizeof(call_t));
            if(call_p == NULL)
            {
                /*
                 * if out of memory then quit
                 */
                say("Out of memory while handling %s: line at line %d\r\n",
                        type, linecount);
                return;
            }

            /*
             * Convert the call to uppercase
             */
            strupr(call);

            /*
             * Store the call
             */
            strcpy(call_p->call, call);

            /*
             * Add to call-list
             */
            call_p->next    = (*call_list_pp);
            (*call_list_pp) = call_p;
        }
        else
        {
            say("Call too long in %s: at line %d, number > 0 expected\r\n", 
                    type, linecount);
        }
        call = strtok(NULL,", \t\r\n");
    }
}

/*
 * Add call(s) from a command from the .ini file to the specified list.
 * This list is used to store destination call signs for one or more specific
 * ports.
 *
 * <token>: 1 AP*,GPS*,DX*,ID*
 *         ^
 *
 * On entry it is assumed that 'strtok' is tokenizing the line and the
 * token is already handled. The tokenizer is now at the point in the
 * string indicated by '^' above (the ':' was changed to '\0' by 'strtok').
 *
 * The first argument is the the text for the type of command - used
 * for error messages. The second argument is a pointer to an array of
 * pointers. The pointer in the array points to the list pointer for each
 * port which pointd to the list on which the calls shall be added.
 *
 * **call_list_pp[] -> *call_list_p[0] -> *call_list (for port 1)
 *                     *call_list_p[1] -> *call_list (for port 2)
 *                     *call_list_p[2] -> *call_list (for port 3)
 *                     *call_list_p[3] -> *call_list (for port 4)
 *                     *call_list_p[4] -> *call_list (for port 5)
 *                     *call_list_p[5] -> *call_list (for port 6)
 *                     *call_list_p[6] -> *call_list (for port 7)
 *                     *call_list_p[7] -> *call_list (for port 8)
 */
void common_add_port_call(char *type, call_t *(*call_list_pp[]))
{
    char       *ports;
    char       *calls;
    char       *tmp;
    short       port;

    ports = strtok(NULL, " \t\r\n");
    if(ports == NULL)
    {
        say("Missing from-port on %s: at line %d\r\n", type, linecount);
        return;
    }

    if(check_ports(ports, "From-port", type, linecount, EXCL_ALLBUT) == 0)
    {
        return;
    }

    /* make an array with the calls */
    calls = strtok(NULL, "\r\n");

    /*
     * digipeat was now setup without problems. Add it to the applicable
     * ports
     */
    port = find_first_port(ports);
    if(port == -1)
    {
        say("From-port out of range (1-%d) on %s: at line %d\r\n",
                mac_ports(), type, linecount);
        return;
    }
    while(port != 0)
    {
        /*
         * Make a copy so strtok can be used to hack out the calls.
         */
        tmp = strdup(calls);
        if(tmp == NULL)
        {
            say("Out of memory while handling %s: line at line %d\r\n",
                    type, linecount);
            return;
        }

        /* add rule to the list (using 0 based port number) */
        common_add_call(tmp, type, &((*call_list_pp)[port - 1]));

        /* throw away out temporary copy */
        free(tmp);

        port = find_next_port();
        if(port == -1)
        {
            say("From-port out of range (1-%d) on %s: at line %d\r\n",
                    mac_ports(), type, linecount);
            return;
        }
    }
}

/*
 * Determine if a call is present in the supplied list
 *
 * Return -1 if the call is the digi's own call
 *         1 if the call is present in the call list
 *         0 if the call is not present
 */
static short common_call_in_list(char *call, call_t* bp)
{
    if(strcmp(call, digi_digi_call) == 0)
    {
        /* found digi's own call */
        return -1;
    }

    while(bp != NULL)
    {
        if(match_call(call, bp->call) != 0)
        {
            /* found call in call-list! */
            return 1;
        }
        bp = bp->next;
    }

    /* Okay, call not in call-list */
    return 0;
}

/*
 * Add call(s) in the owner command from the .ini file to the
 * digi_owner list. This list is used to store call signs that
 * are allowed to do remote maintenance on the node
 *
 * owner: PE1DNN,PE1DNN-7,PE1MEW
 *       ^
 *
 * On entry it is assumed that 'strtok' is tokenizing the line and the
 * 'owner' token is already handled. The tokenizer is now at the point
 * in the string indicated by '^' above (the ':' was changed to '\0' by
 * 'strtok').
 */
void add_owner()
{
    common_add_call(strtok(NULL,"\0"), "owner", &digi_owner_p);
}

/*
 * Determine if a call is allowed to do maintenance
 *
 * Return -1 if the call is the digi's own call
 *         1 if the call is in the owners list
 *         0 if the call is not in the owners list
 */
short is_owner(char *call)
{
    return common_call_in_list(call, digi_owner_p);
}

/*
 * Return the first call in the owner list which is the main owner
 * of the digipeater
 *
 * Return  pointer to call when present
 *         NULL if the owner list is empty
 */
char *digi_owner()
{
    call_t *p;

    if(digi_owner_p == NULL)
    {
        return NULL;
    }

    /*
     * The first defined call is at the end of the list...
     * we have to walk through the list until the end
     */
    p = digi_owner_p;

    while(p->next != NULL)
    {
        p = p->next;
    }

    /* return a pointer to the call */
    return &(p->call[0]);
}

/*
 * Add call(s) in the block command from the .ini file to the digi_block_p
 * list. This list is used to store call signs that are blocked for
 * digipeating via this digipeater (like N0CALL, MYCALL etc.)
 *
 * block: N0CALL,NOCALL,MYCALL
 *       ^
 *
 * On entry it is assumed that 'strtok' is tokenizing the line and the
 * 'block' token is already handled. The tokenizer is now at the point
 * in the string indicated by '^' above (the ':' was changed to '\0' by
 * 'strtok').
 */
void add_block()
{
    common_add_call(strtok(NULL,"\0"), "block", &digi_block_p);
}

/*
 * Determine if a call is blocked
 *
 * Return -1 if the call is the digi's own call
 *         1 if the call is in the blocked list
 *         0 if the call is not blocked
 */
short is_block(char *call)
{
    return common_call_in_list(call, digi_block_p);
}

/*
 * Add call(s) in the block command from the .ini file to the digi_via_block_p
 * list. This list is used to store call signs of digis, packets that passed
 * via that digipeater are blocked for digipeating via this digipeater
 * (like TCPIP, IGATE etc.)
 *
 * via_block: TCPIP,IGATE
 *           ^
 *
 * On entry it is assumed that 'strtok' is tokenizing the line and the
 * 'via_block' token is already handled. The tokenizer is now at the point
 * in the string indicated by '^' above (the ':' was changed to '\0' by
 * 'strtok').
 */
void add_via_block()
{
    common_add_call(strtok(NULL,"\0"), "via_block", &digi_via_block_p);
}

/*
 * Determine if a digipeater path is blocked
 *
 * Return -1 if the digi is the digi's own call
 *         1 if the digi is in the blocked list
 *         0 if the digi is not blocked
 */
short is_via_block(char *call)
{
    return common_call_in_list(call, digi_via_block_p);
}

/*
 * Add call(s) in the msg_block command from the .ini file to the
 * digi_msg_block_p list. This list is used to store call signs that
 * are blocked for querying the digipeater for responses.
 *
 * msg_block: N0CALL,NOCALL,MYCALL
 *           ^
 *
 * On entry it is assumed that 'strtok' is tokenizing the line and the
 * 'block' token is already handled. The tokenizer is now at the point
 * in the string indicated by '^' above (the ':' was changed to '\0' by
 * 'strtok').
 */
void add_msg_block()
{
    common_add_call(strtok(NULL, "\0"), "msg_block", &digi_msg_block_p);
}

/*
 * Determine if a call is blocked for queries/messages.
 *
 * Return -1 if the call is the digi's own call
 *         1 if the call is in the message blocked list
 *         0 if the call is not blocked for messages
 */
short is_msg_block(char *call)
{
    return common_call_in_list(call, digi_msg_block_p);
}

/*
 * Add call(s) in the allow_to command from the .ini file to the
 * digi_allow_to_p array of lists. This list is used to store destination
 * call signs that are allowed in a packet for digipeating via this
 * digipeater (like AP*, GPS*, DX*, ID* etc.)
 *
 * allow_to: AP*,GPS*,DX*,ID*
 *          ^
 *
 * On entry it is assumed that 'strtok' is tokenizing the line and the
 * 'allow_to' token is already handled. The tokenizer is now at the point
 * in the string indicated by '^' above (the ':' was changed to '\0' by
 * 'strtok').
 */
void add_allow_to()
{
    call_t **allow_to_pp;

    allow_to_pp = &(digi_allow_to_p[0]);

    common_add_port_call("allow_to", &allow_to_pp);
}

/*
 * Determine if a desitnation call is allowed
 *
 * Return -1 if the call is the digi's own call
 *         1 if the call is in the allowed_to list
 *         0 if the call is not in the allowed_to list
 */
short is_allowed_to(char *call, short port)
{
    call_t *allow_to_p;

    allow_to_p = digi_allow_to_p[port - 1];

    /* if not specified, allow everybody */
    if(allow_to_p == NULL)
    {
        return 1;
    }

    return common_call_in_list(call, allow_to_p);
}

/*
 * Add call(s) in the allow_from command from the .ini file to the
 * digi_allow_from_p array of lists. This list is used to store originator
 * call signs that are allowed in a packet for digipeating via this
 * digipeater (like only from a specific digipeater, for example PE1MEW-2
 * and PD0JEY-2 etc.)
 *
 * allow_from: 1 PE1MEW-2,PD0JEY-2
 *            ^
 *
 * On entry it is assumed that 'strtok' is tokenizing the line and the
 * 'allow_from' token is already handled. The tokenizer is now at the point
 * in the string indicated by '^' above (the ':' was changed to '\0' by
 * 'strtok').
 */
void add_allow_from()
{
    call_t **digi_allow_from_pp;

    digi_allow_from_pp = &(digi_allow_from_p[0]);

    common_add_port_call("allow_from", &digi_allow_from_pp);
}

/*
 * Determine if a originator call is allowed
 *
 * Return -1 if the call is the digi's own call
 *         1 if the call is in the allowed_from list
 *         0 if the call is not in the allowed_from list
 */
short is_allowed_from(char *call, short port)
{
    call_t *allow_from_p;

    allow_from_p = digi_allow_from_p[port - 1];

    /* if not specified, allow everybody */
    if(allow_from_p == NULL)
    {
        return 1;
    }

    return common_call_in_list(call, allow_from_p);
}
